# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=invalid-name, unnecessary-lambda-assignment

r"""
Solver classes.
"""


from typing import Optional, Union, Tuple, Any, Type, List, Callable
from warnings import warn

import numpy as np

from scipy.integrate._ivp.ivp import OdeResult

from qiskit import QiskitError
from qiskit.pulse import Schedule, ScheduleBlock
from qiskit.pulse.transforms import block_to_schedule

from qiskit.circuit import Gate, QuantumCircuit
from qiskit.quantum_info.operators.base_operator import BaseOperator
from qiskit.quantum_info.operators.channel.quantum_channel import QuantumChannel
from qiskit.quantum_info.states.quantum_state import QuantumState
from qiskit.quantum_info import SuperOp, Operator, DensityMatrix

from qiskit_dynamics import ArrayLike
from qiskit_dynamics import DYNAMICS_NUMPY as unp
from qiskit_dynamics import DYNAMICS_NUMPY_ALIAS as numpy_alias
from qiskit_dynamics.arraylias.alias import _isArrayLike

from qiskit_dynamics.models import (
    HamiltonianModel,
    LindbladModel,
    RotatingFrame,
    rotating_wave_approximation,
)
from qiskit_dynamics.signals import Signal, DiscreteSignal, SignalList
from qiskit_dynamics.pulse import InstructionToSignals

from .solver_functions import solve_lmde, _is_diffrax_method
from .solver_utils import (
    is_lindblad_model_vectorized,
    is_lindblad_model_not_vectorized,
    setup_args_lists,
)


try:
    from jax import core, jit
    import jax.numpy as jnp
except ImportError:
    pass


class Solver:
    r"""Solver class for simulating both Hamiltonian and Lindblad dynamics, with high level
    type-handling of input states.

    If only Hamiltonian information is provided, this class will internally construct a
    :class:`.HamiltonianModel` instance, and simulate the model using the Schrodinger equation
    :math:`\dot{y}(t) = -iH(t)y(t)` (see the :meth:`.Solver.solve` method documentation for details
    on how different initial state types are handled). :class:`.HamiltonianModel` represents a
    decomposition of the Hamiltonian of the form:

    .. math::

        H(t) = H_0 + \sum_i s_i(t) H_i,

    where :math:`H_0` is the static component, the :math:`H_i` are the time-dependent components of
    the Hamiltonian, and the :math:`s_i(t)` are the time-dependent signals, specifiable as either
    :class:`.Signal` objects, or constructed from Qiskit Pulse schedules if :class:`.Solver` is
    configured for Pulse simulation (see below).

    If dissipators are specified as part of the model, then a :class:`.LindbladModel` is
    constructed, and simulations are performed by solving the Lindblad equation:

    .. math::

        \dot{y}(t) = -i[H(t), y(t)] + \mathcal{D}_0(y(t)) + \mathcal{D}(t)(y(t)),

    where :math:`H(t)` is the Hamiltonian part, specified as above, and :math:`\mathcal{D}_0` and
    :math:`\mathcal{D}(t)` are the static and time-dependent portions of the dissipator, given by:

    .. math::

        \mathcal{D}_0(y(t)) = \sum_j N_j y(t) N_j^\dagger
                                      - \frac{1}{2} \{N_j^\dagger N_j, y(t)\},

    and

    .. math::

        \mathcal{D}(t)(y(t)) = \sum_j \gamma_j(t) L_j y(t) L_j^\dagger
                                  - \frac{1}{2} \{L_j^\dagger L_j, y(t)\},

    with :math:`N_j` the static dissipators, :math:`L_j` the time-dependent dissipator operators,
    and :math:`\gamma_j(t)` the time-dependent signals specifiable as either :class:`.Signal`
    objects, or constructed from Qiskit Pulse schedules if :class:`.Solver` is configured for Pulse
    simulation (see below).

    Transformations on the model can be specified via the optional arguments:

    * ``rotating_frame``: Transforms the model into a rotating frame. Note that the operator
      specifying the frame will be substracted from the ``static_hamiltonian``. If supplied as a 1d
      array, ``rotating_frame`` is interpreted as the diagonal elements of a diagonal matrix. Given
      a frame operator :math:`F = -i H_0`, for the Schrodinger equation entering the rotating frame
      of :math:`F`, corresponds to transforming the solution as :math:`y(t) \mapsto exp(-tF)y(t)`,
      and for the Lindblad equation it corresponds to transforming the solution as
      :math:`y(t) \mapsto exp(-tF)y(t)exp(tF)`. See :class:`.RotatingFrame` for more details.
    * ``in_frame_basis``: Whether to represent the model in the basis in which the frame operator is
      diagonal, henceforth called the "frame basis". If ``rotating_frame`` is ``None`` or was
      supplied as a 1d array, this kwarg has no effect. If ``rotating_frame`` was specified as a 2d
      array, the frame basis is the diagonalizing basis supplied by ``np.linalg.eigh``. If
      ``in_frame_basis==True``, this objects behaves as if all operators were supplied in the frame
      basis: calls to ``solve`` will assume the initial state is supplied in the frame basis, and
      the results will be returned in the frame basis. If ``in_frame_basis==False``, the system will
      still be solved in the frame basis for efficiency, however the initial state (and final output
      states) will automatically be transformed into (and, respectively, out of) the frame basis.
    * ``rwa_cutoff_freq`` and ``rwa_carrier_freqs``: Performs a rotating wave approximation (RWA) on
      the model with cutoff frequency ``rwa_cutoff_freq``, assuming the time-dependent coefficients
      of the model have carrier frequencies specified by ``rwa_carrier_freqs``. If
      ``dissipator_operators is None``, ``rwa_carrier_freqs`` must be a list of floats of length
      equal to ``hamiltonian_operators``, and if ``dissipator_operators is not None``,
      ``rwa_carrier_freqs`` must be a ``tuple`` of lists of floats, with the first entry the list of
      carrier frequencies for ``hamiltonian_operators``, and the second entry the list of carrier
      frequencies for ``dissipator_operators``. See :func:`.rotating_wave_approximation` for details
      on the mathematical approximation.

    Additionally, the ``array_library`` argument controls the underlying array representation used
    to store and evaluate the model. See the
    :ref:`model evaluation section of the Models API documentation <model evaluation>`
    for a more detailed description of this argument.

      .. note::
            When using the ``rwa_cutoff_freq`` optional argument, :class:`.Solver` cannot be
            instantiated within a JAX-transformable function. However, after construction, instances
            can still be used within JAX-transformable functions regardless of whether an
            ``rwa_cutoff_freq`` is set.

    :class:`.Solver` can be configured to simulate Qiskit Pulse schedules by setting all of the
    following parameters, which determine how Pulse schedules are interpreted:

    * ``hamiltonian_channels``: List of channels in string format corresponding to the
      time-dependent coefficients of ``hamiltonian_operators``.
    * ``dissipator_channels``: List of channels in string format corresponding to time-dependent
      coefficients of ``dissipator_operators``.
    * ``channel_carrier_freqs``: Dictionary mapping channel names to frequencies. A frequency
      must be specified for every channel appearing in ``hamiltonian_channels`` and
      ``dissipator_channels``. When simulating ``schedule``\s, these frequencies are interpreted as
      the analog carrier frequencies associated with the channel; deviations from these frequencies
      due to ``SetFrequency`` or ``ShiftFrequency`` instructions are implemented by digitally
      modulating the samples for the channel envelope. If an ``rwa_cutoff_freq`` is specified, and
      no ``rwa_carrier_freqs`` is specified, these frequencies will be used for the RWA.
    * ``dt``: The envelope sample width.

    If configured to simulate Pulse schedules, and a JAX-based solver method is chosen when calling
    :meth:`.Solver.solve`, :meth:`.Solver.solve` will automatically attempt to compile a single
    function to re-use for all schedule simulations.

    The evolution given by the model can be simulated by calling :meth:`.Solver.solve`, which calls
    :func:`.solve_lmde`, and does various automatic type handling operations for
    :mod:`qiskit.quantum_info` state and super operator types.
    """

    def __init__(
        self,
        static_hamiltonian: Optional[ArrayLike] = None,
        hamiltonian_operators: Optional[ArrayLike] = None,
        static_dissipators: Optional[ArrayLike] = None,
        dissipator_operators: Optional[ArrayLike] = None,
        hamiltonian_channels: Optional[List[str]] = None,
        dissipator_channels: Optional[List[str]] = None,
        channel_carrier_freqs: Optional[dict] = None,
        dt: Optional[float] = None,
        rotating_frame: Optional[Union[ArrayLike, RotatingFrame]] = None,
        in_frame_basis: bool = False,
        array_library: Optional[str] = None,
        vectorized: Optional[bool] = None,
        rwa_cutoff_freq: Optional[float] = None,
        rwa_carrier_freqs: Optional[Union[ArrayLike, Tuple[ArrayLike, ArrayLike]]] = None,
        validate: bool = True,
    ):
        """Initialize solver with model information.

        Args:
            static_hamiltonian: Constant Hamiltonian term. If a ``rotating_frame``
                is specified, the ``frame_operator`` will be subtracted from the static_hamiltonian.
            hamiltonian_operators: Hamiltonian operators.
            static_dissipators: Constant dissipation operators.
            dissipator_operators: Dissipation operators with time-dependent coefficients.
            hamiltonian_channels: List of channel names in pulse schedules corresponding to
                Hamiltonian operators.
            dissipator_channels: List of channel names in pulse schedules corresponding to
                dissipator operators.
            channel_carrier_freqs: Dictionary mapping channel names to floats which represent the
                carrier frequency of the pulse channel with the corresponding name.
            dt: Sample rate for simulating pulse schedules.
            rotating_frame: Rotating frame to transform the model into. Rotating frames which are
                diagonal can be supplied as a 1d array of the diagonal elements, to explicitly
                indicate that they are diagonal.
            in_frame_basis: Whether to represent the model in the basis in which the rotating
                frame operator is diagonalized. See class documentation for a more detailed
                explanation on how this argument affects object behaviour.
            array_library: Array library to use for storing operators of underlying model. See the
                :ref:`model evaluation section of the Models API documentation <model evaluation>`
                for a more detailed description of this argument.
            vectorized: If including dissipator terms, whether or not to construct the
                :class:`.LindbladModel` in vectorized form. See the
                :ref:`model evaluation section of the Models API documentation <model evaluation>`
                for a more detailed description of this argument.
            rwa_cutoff_freq: Rotating wave approximation cutoff frequency. If ``None``, no
                approximation is made.
            rwa_carrier_freqs: Carrier frequencies to use for rotating wave approximation.
                If no time dependent coefficients in model leave as ``None``, if no time-dependent
                dissipators specify as a list of frequencies for each Hamiltonian operator, and if
                time-dependent dissipators present specify as a tuple of lists of frequencies, one
                for Hamiltonian operators and one for dissipators.
            validate: Whether or not to validate Hamiltonian operators as being Hermitian.

        Raises:
            QiskitError: If arguments concerning pulse-schedule interpretation are insufficiently
            specified.
        """

        # set pulse specific information if specified
        self._hamiltonian_channels = None
        self._dissipator_channels = None
        self._all_channels = None
        self._channel_carrier_freqs = None
        self._dt = None
        self._schedule_converter = None

        if any([dt, channel_carrier_freqs, hamiltonian_channels, dissipator_channels]):
            all_channels = []

            if hamiltonian_channels is not None:
                hamiltonian_channels = [chan.lower() for chan in hamiltonian_channels]
                if hamiltonian_operators is None or len(hamiltonian_operators) != len(
                    hamiltonian_channels
                ):
                    raise QiskitError(
                        """hamiltonian_channels must have same length as hamiltonian_operators"""
                    )
                for chan in hamiltonian_channels:
                    if chan not in all_channels:
                        all_channels.append(chan)

            self._hamiltonian_channels = hamiltonian_channels

            if dissipator_channels is not None:
                dissipator_channels = [chan.lower() for chan in dissipator_channels]
                for chan in dissipator_channels:
                    if chan not in all_channels:
                        all_channels.append(chan)
                if dissipator_operators is None or len(dissipator_operators) != len(
                    dissipator_channels
                ):
                    raise QiskitError(
                        """dissipator_channels must have same length as dissipator_operators"""
                    )

            self._dissipator_channels = dissipator_channels
            self._all_channels = all_channels

            if channel_carrier_freqs is None:
                channel_carrier_freqs = {}
            else:
                channel_carrier_freqs = {
                    key.lower(): val for key, val in channel_carrier_freqs.items()
                }

            for chan in all_channels:
                if chan not in channel_carrier_freqs:
                    raise QiskitError(
                        f"""Channel '{chan}' does not have carrier frequency specified in
                        channel_carrier_freqs."""
                    )

            if len(channel_carrier_freqs) == 0:
                channel_carrier_freqs = None
            self._channel_carrier_freqs = channel_carrier_freqs

            if dt is not None:
                self._dt = dt

                self._schedule_converter = InstructionToSignals(
                    dt=self._dt, carriers=self._channel_carrier_freqs, channels=self._all_channels
                )
            else:
                raise QiskitError("dt must be specified if channel information is provided.")

        # setup model
        model = None
        if static_dissipators is None and dissipator_operators is None:
            model = HamiltonianModel(
                static_operator=static_hamiltonian,
                operators=hamiltonian_operators,
                rotating_frame=rotating_frame,
                in_frame_basis=in_frame_basis,
                array_library=array_library,
                validate=validate,
            )
        else:
            model = LindbladModel(
                static_hamiltonian=static_hamiltonian,
                hamiltonian_operators=hamiltonian_operators,
                static_dissipators=static_dissipators,
                dissipator_operators=dissipator_operators,
                rotating_frame=rotating_frame,
                in_frame_basis=in_frame_basis,
                array_library=array_library,
                vectorized=vectorized,
                validate=validate,
            )

        self._rwa_signal_map = None
        self._model = model
        if rwa_cutoff_freq:
            # if rwa_carrier_freqs is None, take from channel_carrier_freqs or set all to 0.
            if rwa_carrier_freqs is None:
                if self._channel_carrier_freqs is not None:
                    if self._hamiltonian_channels is not None:
                        rwa_carrier_freqs = [
                            self._channel_carrier_freqs[c] for c in self._hamiltonian_channels
                        ]

                    if self._dissipator_channels is not None:
                        rwa_carrier_freqs = (
                            rwa_carrier_freqs,
                            [self._channel_carrier_freqs[c] for c in self._dissipator_channels],
                        )
                else:
                    rwa_carrier_freqs = []
                    if hamiltonian_operators is not None:
                        rwa_carrier_freqs = [0.0] * len(hamiltonian_operators)

                    if dissipator_operators is not None:
                        rwa_carrier_freqs = (rwa_carrier_freqs, [0.0] * len(dissipator_operators))

            if isinstance(rwa_carrier_freqs, tuple):
                rwa_ham_sigs = None
                rwa_lindblad_sigs = None
                if rwa_carrier_freqs[0]:
                    rwa_ham_sigs = [Signal(1.0, carrier_freq=freq) for freq in rwa_carrier_freqs[0]]
                if rwa_carrier_freqs[1]:
                    rwa_lindblad_sigs = [
                        Signal(1.0, carrier_freq=freq) for freq in rwa_carrier_freqs[1]
                    ]

                self._model.signals = (rwa_ham_sigs, rwa_lindblad_sigs)
            else:
                rwa_sigs = [Signal(1.0, carrier_freq=freq) for freq in rwa_carrier_freqs]

                if isinstance(model, LindbladModel):
                    rwa_sigs = (rwa_sigs, None)

                self._model.signals = rwa_sigs

            self._model, rwa_signal_map = rotating_wave_approximation(
                self._model, rwa_cutoff_freq, return_signal_map=True
            )
            self._rwa_signal_map = rwa_signal_map

            # clear signals
            self._set_new_signals(None)

    @property
    def model(self) -> Union[HamiltonianModel, LindbladModel]:
        """The model of the system, either a Hamiltonian or Lindblad model."""
        return self._model

    def solve(
        self,
        t_span: ArrayLike,
        y0: Union[ArrayLike, QuantumState, BaseOperator],
        signals: Optional[
            Union[
                List[Union[Schedule, ScheduleBlock]],
                List[Signal],
                Tuple[List[Signal], List[Signal]],
            ]
        ] = None,
        convert_results: bool = True,
        **kwargs,
    ) -> Union[OdeResult, List[OdeResult]]:
        r"""Solve a dynamical problem, or a set of dynamical problems.

        Calls :func:`.solve_lmde`, and returns an ``OdeResult``
        object in the style of ``scipy.integrate.solve_ivp``, with results
        formatted to be the same types as the input. See Additional Information
        for special handling of various input types, and for specifying multiple
        simulations at once.

        Args:
            t_span: Time interval to integrate over.
            y0: Initial state.
            signals: Specification of time-dependent coefficients to simulate, either in
                Signal format or as Qiskit Pulse Pulse schedules. If specifying in Signal format, if
                ``dissipator_operators is None``, specify as a list of signals for the Hamiltonian
                component, otherwise specify as a tuple of two lists, one for Hamiltonian
                components, and one for the ``dissipator_operators`` coefficients.
            convert_results: If ``True``, convert returned solver state results to the same class
                as ``y0``. If ``False``, states will be returned in the native array type used by
                the specified solver method.
            **kwargs: Keyword args passed to :func:`.solve_lmde`.

        Returns:
            OdeResult: object with formatted output types.

        Raises:
            QiskitError: Initial state ``y0`` is of invalid shape. If ``signals`` specifies
                ``Schedule`` simulation but ``Solver`` hasn't been configured to simulate pulse
                schedules.

        Additional Information:

        The behaviour of this method is impacted by the input type of ``y0`` and the internal model,
        summarized in the following table:

        .. list-table:: Type-based behaviour
           :widths: 10 10 10 70
           :header-rows: 1

           * - ``y0`` type
             - Model type
             - ``yf`` type
             - Description
           * - ``ArrayLike``, ``np.ndarray``, ``Operator``
             - Any
             - Same as ``y0``
             - Solves either the Schrodinger equation or Lindblad equation
               with initial state ``y0`` as specified.
           * - ``Statevector``
             - ``HamiltonianModel``
             - ``Statevector``
             - Solves the Schrodinger equation with initial state ``y0``.
           * - ``DensityMatrix``
             - ``HamiltonianModel``
             - ``DensityMatrix``
             - Solves the Schrodinger equation with initial state the identity matrix to compute
               the unitary, then conjugates ``y0`` with the result to solve for the density matrix.
           * - ``Statevector``, ``DensityMatrix``
             - ``LindbladModel``
             - ``DensityMatrix``
             - Solve the Lindblad equation with initial state ``y0``, converting to a
               ``DensityMatrix`` first if ``y0`` is a ``Statevector``.
           * - ``QuantumChannel``
             - ``HamiltonianModel``
             - ``SuperOp``
             - Converts ``y0`` to a ``SuperOp`` representation, then solves the Schrodinger
               equation with initial state the identity matrix to compute the unitary and
               composes with ``y0``.
           * - ``QuantumChannel``
             - ``LindbladModel``
             - ``SuperOp``
             - Solves the vectorized Lindblad equation with initial state ``y0``. ``vectorized``
               must be set to ``True``.

        In some cases (e.g. if using JAX), wrapping the returned states in the type given in the
        ``yf`` type column above may be undesirable. Setting ``convert_results=False`` prevents this
        wrapping, while still allowing usage of the automatic type-handling for the input state.

        In addition to the above, this method can be used to specify multiple simulations
        simultaneously. This can be done by specifying one or more of the arguments ``t_span``,
        ``y0``, or ``signals`` as a list of valid inputs. For this mode of operation, all of these
        arguments must be either lists of the same length, or a single valid input, which will be
        used repeatedly.

        For example the following code runs three simulations, returning results in a list:

        .. code-block:: python

            t_span = [span1, span2, span3]
            y0 = [state1, state2, state3]
            signals = [signals1, signals2, signals3]

            results = solver.solve(t_span=t_span, y0=y0, signals=signals)

        The following code block runs three simulations, for different sets of signals, repeatedly
        using the same ``t_span`` and ``y0``:

        .. code-block:: python

            t_span = [t0, tf]
            y0 = state1
            signals = [signals1, signals2, signal3]

            results = solver.solve(t_span=t_span, y0=y0, signals=signals)
        """

        # convert any ScheduleBlocks to Schedules
        if isinstance(signals, ScheduleBlock):
            signals = block_to_schedule(signals)
        elif isinstance(signals, list):
            signals = [block_to_schedule(x) if isinstance(x, ScheduleBlock) else x for x in signals]

        # validate and setup list of simulations
        [t_span_list, y0_list, signals_list], multiple_sims = setup_args_lists(
            args_list=[t_span, y0, signals],
            args_names=["t_span", "y0", "signals"],
            args_to_list=[t_span_to_list, _y0_to_list, _signals_to_list],
        )

        all_results = None
        method = kwargs.get("method", "")
        if (
            (method == "jax_odeint" or _is_diffrax_method(method))
            and all(isinstance(x, Schedule) for x in signals_list)
            # check if jit transformation is already performed.
            and not (isinstance(jnp.array(0), core.Tracer))
        ):
            if self.model.array_library not in ["numpy", "jax", "jax_sparse"]:
                warn(
                    "Attempting to internally JAX-compile simulation of schedules, with "
                    'Solver.model.array_library not in ["numpy", "jax", "jax_sparse"]. If an error '
                    "is not raised, explicitly set array_library at Solver instantation to one of "
                    "these options to remove this warning."
                )

            all_results = self._solve_schedule_list_jax(
                t_span_list=t_span_list,
                y0_list=y0_list,
                schedule_list=signals_list,
                convert_results=convert_results,
                **kwargs,
            )
        else:
            all_results = self._solve_list(
                t_span_list=t_span_list,
                y0_list=y0_list,
                signals_list=signals_list,
                convert_results=convert_results,
                **kwargs,
            )

        # ensure model signals are empty
        self._set_new_signals(None)

        if multiple_sims is False:
            return all_results[0]

        return all_results

    def _solve_list(
        self,
        t_span_list: List[ArrayLike],
        y0_list: List[Union[ArrayLike, QuantumState, BaseOperator]],
        signals_list: Optional[
            Union[List[Schedule], List[List[Signal]], List[Tuple[List[Signal], List[Signal]]]]
        ] = None,
        convert_results: bool = True,
        **kwargs,
    ) -> List[OdeResult]:
        """Run a list of simulations."""

        all_results = []
        for t_span, y0, signals in zip(t_span_list, y0_list, signals_list):
            if isinstance(signals, Schedule):
                signals = self._schedule_to_signals(signals)

            self._set_new_signals(signals)

            # setup initial state
            y0, y0_input, y0_cls, state_type_wrapper = validate_and_format_initial_state(
                y0, self.model
            )

            results = solve_lmde(generator=self.model, t_span=t_span, y0=y0, **kwargs)
            results.y = format_final_states(results.y, self.model, y0_input, y0_cls)

            if y0_cls is not None and convert_results:
                results.y = [state_type_wrapper(yi) for yi in results.y]

            all_results.append(results)

        self._set_new_signals(None)

        return all_results

    def _solve_schedule_list_jax(
        self,
        t_span_list: List[ArrayLike],
        y0_list: List[Union[ArrayLike, QuantumState, BaseOperator]],
        schedule_list: List[Schedule],
        convert_results: bool = True,
        **kwargs,
    ) -> List[OdeResult]:
        """Run a list of schedule simulations utilizing JAX compilation.
        The jitting strategy is to define a function whose inputs are t_span, y0 as an array, the
        samples for all channels in a single large array, and other initial state information. To
        avoid recompilation for schedules with a different number of samples, i.e. a different
        duration, all schedules are padded to be the length of the schedule with the max duration.
        """

        # determine fixed array shape for containing all samples
        max_duration = 0
        for idx, sched in enumerate(schedule_list):
            max_duration = max(sched.duration, max_duration)
        all_samples_shape = (len(self._all_channels), max_duration)

        # define sim function to jit
        def sim_function(t_span, y0, all_samples, y0_input, y0_cls):
            # store signals to ensure purity
            model_sigs = self.model.signals

            # re-construct signals from the samples
            signals = []
            for idx, samples in enumerate(all_samples):
                carrier_freq = self._channel_carrier_freqs[self._all_channels[idx]]
                signals.append(
                    DiscreteSignal(dt=self._dt, samples=samples, carrier_freq=carrier_freq)
                )

            # map signals to correct structure for model
            signals = organize_signals_to_channels(
                signals,
                self._all_channels,
                self.model.__class__,
                self._hamiltonian_channels,
                self._dissipator_channels,
            )

            self._set_new_signals(signals)

            results = solve_lmde(generator=self.model, t_span=t_span, y0=y0, **kwargs)
            results.y = format_final_states(results.y, self.model, y0_input, y0_cls)

            # reset signals to ensure purity
            self.model.signals = model_sigs

            return results.t, results.y

        jit_sim_function = jit(sim_function, static_argnums=(4,))

        # run simulations
        all_results = []
        for t_span, y0, sched in zip(t_span_list, y0_list, schedule_list):
            # setup initial state
            y0, y0_input, y0_cls, state_type_wrapper = validate_and_format_initial_state(
                y0, self.model
            )

            # setup array of all samples
            all_signals = self._schedule_converter.get_signals(sched)

            all_samples = np.zeros(all_samples_shape, dtype=complex)
            for idx, sig in enumerate(all_signals):
                all_samples[idx, 0 : len(sig.samples)] = np.array(sig.samples)

            results_t, results_y = jit_sim_function(
                unp.asarray(t_span),
                unp.asarray(y0),
                unp.asarray(all_samples),
                unp.asarray(y0_input),
                y0_cls,
            )
            results = OdeResult(t=results_t, y=results_y)

            if y0_cls is not None and convert_results:
                results.y = [state_type_wrapper(yi) for yi in results.y]

            all_results.append(results)

        return all_results

    def _set_new_signals(self, signals):
        """Helper function for setting new signals in self.model."""
        if signals is not None:
            # if Lindblad model and signals are given as a list set as Hamiltonian part of signals
            if isinstance(self.model, LindbladModel) and isinstance(signals, (list, SignalList)):
                signals = (signals, None)

            if self._rwa_signal_map:
                signals = self._rwa_signal_map(signals)
            self.model.signals = signals
        else:
            if isinstance(self.model, LindbladModel):
                self.model.signals = (None, None)
            else:
                self.model.signals = None

    def _schedule_to_signals(self, schedule: Schedule):
        """Convert a schedule into the signal format required by the model."""
        if self._schedule_converter is None:
            raise QiskitError("Solver instance not configured for pulse Schedule simulation.")

        return organize_signals_to_channels(
            self._schedule_converter.get_signals(schedule),
            self._all_channels,
            self.model.__class__,
            self._hamiltonian_channels,
            self._dissipator_channels,
        )


def initial_state_converter(obj: Any) -> Tuple[ArrayLike, Type, Callable]:
    """Convert initial state object to an Array, the type of the initial input, and return
    function for constructing a state of the same type.

    Args:
        obj: An initial state.

    Returns:
        tuple: (Array, Type, Callable)
    """
    # pylint: disable=invalid-name
    y0_cls = None
    if _isArrayLike(obj):
        y0, y0_cls, wrapper = obj, None, lambda x: x
    if isinstance(obj, QuantumState):
        y0, y0_cls = obj.data, obj.__class__
        wrapper = lambda x: y0_cls(np.array(x), dims=obj.dims())
    elif isinstance(obj, QuantumChannel):
        y0, y0_cls = SuperOp(obj).data, SuperOp
        wrapper = lambda x: SuperOp(
            np.array(x), input_dims=obj.input_dims(), output_dims=obj.output_dims()
        )
    elif isinstance(obj, (BaseOperator, Gate, QuantumCircuit)):
        y0, y0_cls = obj.data, Operator
        wrapper = lambda x: Operator(
            np.array(x), input_dims=obj.input_dims(), output_dims=obj.output_dims()
        )
    else:
        y0, y0_cls, wrapper = unp.asarray(obj), None, lambda x: x

    return y0, y0_cls, wrapper


def validate_and_format_initial_state(y0: any, model: Union[HamiltonianModel, LindbladModel]):
    """Format initial state for simulation. This function encodes the logic of how simulations are
    run based on initial state type.

    Args:
        y0: The user-specified input state.
        model: The model contained in the solver.

    Returns:
        Tuple containing the input state to pass to the solver, the user-specified input as an
        array, the class of the user specified input, and a function for converting the output
        states to the right class.

    Raises:
        QiskitError: Initial state ``y0`` is of invalid shape relative to the model.
    """

    if isinstance(y0, QuantumState) and isinstance(model, LindbladModel):
        y0 = DensityMatrix(y0)

    y0, y0_cls, wrapper = initial_state_converter(y0)

    y0_input = y0

    # validate types
    if (y0_cls is SuperOp) and is_lindblad_model_not_vectorized(model):
        raise QiskitError(
            """Simulating SuperOp for a LindbladModel requires setting vectorized evaluation.
            Set vectorized=True when constructing LindbladModel.
            """
        )

    # if Simulating density matrix or SuperOp with a HamiltonianModel, simulate the unitary
    if y0_cls in [DensityMatrix, SuperOp] and isinstance(model, HamiltonianModel):
        y0 = np.eye(model.dim, dtype=complex)
    # if LindbladModel is vectorized and simulating a density matrix, flatten
    elif (y0_cls is DensityMatrix) and is_lindblad_model_vectorized(model):
        y0 = y0.flatten(order="F")

    # validate y0 shape before passing to solve_lmde
    if isinstance(model, HamiltonianModel) and (y0.shape[0] != model.dim or y0.ndim > 2):
        raise QiskitError("""Shape mismatch for initial state y0 and HamiltonianModel.""")
    if is_lindblad_model_vectorized(model) and (y0.shape[0] != model.dim**2 or y0.ndim > 2):
        raise QiskitError(
            """Shape mismatch for initial state y0 and LindbladModel
                             in vectorized evaluation mode."""
        )
    if is_lindblad_model_not_vectorized(model) and y0.shape[-2:] != (
        model.dim,
        model.dim,
    ):
        raise QiskitError("""Shape mismatch for initial state y0 and LindbladModel.""")

    return y0, y0_input, y0_cls, wrapper


def format_final_states(y, model, y0_input, y0_cls):
    """Format final states for a single simulation."""

    y = unp.asarray(y)

    if y0_cls is DensityMatrix and isinstance(model, HamiltonianModel):
        # conjugate by unitary
        return y @ y0_input @ y.conj().transpose((0, 2, 1))
    elif y0_cls is SuperOp and isinstance(model, HamiltonianModel):
        # convert to SuperOp and compose
        return (
            numpy_alias(like=y)
            .einsum("nka,nlb->nklab", y.conj(), y)
            .reshape(y.shape[0], y.shape[1] ** 2, y.shape[1] ** 2)
            @ y0_input
        )
    elif (y0_cls is DensityMatrix) and is_lindblad_model_vectorized(model):
        return y.reshape((len(y),) + y0_input.shape, order="F")

    return y


def t_span_to_list(t_span):
    """Check if t_span is validly specified as a single interval or a list of intervals, and return
    as a list in either case."""
    was_list = False
    t_span_ndim = _nested_ndim(t_span)
    if t_span_ndim > 2:
        raise QiskitError("t_span must be either 1d or 2d.")
    if t_span_ndim == 1:
        t_span = [t_span]
    else:
        was_list = True

    return t_span, was_list


def _y0_to_list(y0):
    """Check if y0 is validly specified as a single initial state or a list of initial states, and
    return as a list in either case."""
    was_list = False
    if not isinstance(y0, list):
        y0 = [y0]
    else:
        was_list = True

    return y0, was_list


def _signals_to_list(signals):
    """Check if signals is validly specified as a single signal specification or a list of such
    specifications, and return as a list in either case."""
    was_list = False
    if signals is None:
        signals = [signals]
    elif isinstance(signals, tuple):
        # single Lindblad
        signals = [signals]
    elif isinstance(signals, list) and isinstance(signals[0], tuple):
        # multiple lindblad
        was_list = True
    elif isinstance(signals, Schedule):
        # pulse simulation
        signals = [signals]
    elif isinstance(signals, list) and isinstance(signals[0], Schedule):
        # multiple pulse simulation
        was_list = True
    elif isinstance(signals, list) and isinstance(signals[0], (list, SignalList)):
        # multiple Hamiltonian signals lists
        was_list = True
    elif isinstance(signals, SignalList) or (
        isinstance(signals, list) and not isinstance(signals[0], (list, SignalList))
    ):
        # single Hamiltonian signals list
        signals = [signals]
    else:
        raise QiskitError("Signals specified in invalid format.")

    return signals, was_list


def organize_signals_to_channels(
    all_signals, all_channels, model_class, hamiltonian_channels, dissipator_channels
):
    """Restructures a list of signals with order corresponding to all_channels, into the correctly
    formatted data structure to pass into model.signals, according to the ordering specified by
    hamiltonian_channels and dissipator_channels.
    """

    if model_class == HamiltonianModel:
        if hamiltonian_channels is not None:
            return [all_signals[all_channels.index(chan)] for chan in hamiltonian_channels]
        else:
            return None
    else:
        hamiltonian_signals = None
        dissipator_signals = None
        if hamiltonian_channels is not None:
            hamiltonian_signals = [
                all_signals[all_channels.index(chan)] for chan in hamiltonian_channels
            ]
        if dissipator_channels is not None:
            dissipator_signals = [
                all_signals[all_channels.index(chan)] for chan in dissipator_channels
            ]
        return (hamiltonian_signals, dissipator_signals)


def _nested_ndim(x):
    """Determine the 'ndim' of x, which could be composed of nested lists and array types."""
    if isinstance(x, (list, tuple)):
        return 1 + _nested_ndim(x[0])
    elif hasattr(x, "ndim"):
        return x.ndim

    # assume scalar
    return 0



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=invalid-name

"""
Module containing Lanczos diagonalization and time evolution algorithms.
"""

from typing import Union, Optional
import numpy as np
from scipy.sparse import csr_matrix

from qiskit_dynamics.arraylias import requires_array_library

try:
    import jax.numpy as jnp
    from jax.lax import scan, cond
except ImportError:
    pass


def lanczos_basis(A: Union[csr_matrix, np.ndarray], y0: np.ndarray, k_dim: int):
    """Tridiagonalises a hermitian array in a Krylov subspace of dimension ``k_dim``
    using Lanczos algorithm.
    reference: https://tensornetwork.org/mps/algorithms/timeevo/global-krylov.html

    Args:
        A : Array to tridiagonalise. Must be hermitian.
        y0 : Vector to initialise Lanczos iteration.
        k_dim : Dimension of the Krylov subspace.

    Returns:
        tridiagonal : Tridiagonal projection of ``A``.
        q_basis : Basis of the Krylov subspace.
    """

    data_type = np.result_type(A.dtype, y0.dtype)
    y0 = np.array(y0).reshape(-1, 1)
    array_dim = A.shape[0]
    q_basis = np.zeros((k_dim, array_dim), dtype=data_type)

    v_p = np.zeros_like(y0)
    projection = np.zeros_like(y0)

    beta = np.zeros((k_dim,), dtype=data_type)
    alpha = np.zeros((k_dim,), dtype=data_type)

    q_basis[[0], :] = y0.T
    projection = A @ y0
    alpha[0] = np.sum(y0.conj() * projection)
    projection = projection - alpha[0] * y0
    beta[0] = np.linalg.norm(projection)

    error = np.finfo(np.float64).eps

    for i in range(1, k_dim, 1):
        if beta[i - 1] < error:
            k_dim = i
            break

        v_p = q_basis[i - 1, :]

        q_basis[[i], :] = projection.T / beta[i - 1]
        projection = A @ q_basis[i, :]
        alpha[i] = q_basis[i, :].conj().T @ projection
        projection = projection - alpha[i] * q_basis[i, :] - beta[i - 1] * v_p
        beta[i] = np.linalg.norm(projection)

        # additional steps to increase accuracy
        delta = q_basis[i, :].conj().T @ projection
        projection -= delta * q_basis[i, :]
        alpha[i] += delta

    tridiagonal = (
        np.diag(alpha[:k_dim], k=0)
        + np.diag(beta[: k_dim - 1], k=-1)
        + np.diag(beta[: k_dim - 1], k=1)
    )
    q_basis = q_basis[:k_dim]
    q_basis = q_basis.T
    return tridiagonal, q_basis


def lanczos_eigh(A: Union[csr_matrix, np.ndarray], y0: np.ndarray, k_dim: int):
    """Finds the lowest (Algebraic) ``k_dim`` eigenvalues and corresponding eigenvectors of a
    hermitian array using Lanczos algorithm.
    Args:
        A : Array to diagonalize. Must be hermitian.
        y0 : Vector to initialise Lanczos iteration.
        k_dim : Dimension of the Krylov subspace.

    Returns:
        q_basis : Basis of the Krylov subspace.
        eigen_values : Lowest ``k_dim`` eigenvalues.
        eigen_vectors_t : Eigenvectors in Krylov-space.
    """

    tridiagonal, q_basis = lanczos_basis(A, y0, k_dim)
    eigen_values, eigen_vectors_t = np.linalg.eigh(tridiagonal)

    return q_basis, eigen_values, eigen_vectors_t


def lanczos_expm(
    A: Union[csr_matrix, np.ndarray],
    y0: np.ndarray,
    k_dim: int,
    scale_factor: Optional[float] = 1,
):
    """Calculates action of matrix exponential of an anti-hermitian array on the state using
    Lanczos algorithm.

    Args:
        A : Array to exponentiate. Must be anti-hermitian.
        y0 : Initial state.
        k_dim : Dimension of the Krylov subspace.
        scale_factor : Maximum step size.

    Returns:
        y_dt : Action of matrix exponential on state.
    """

    if y0.ndim == 1:
        A = 1j * A  # make hermitian
        y0_norm = np.linalg.norm(y0)
        q_basis, eigen_values, eigen_vectors_t = lanczos_eigh(A, y0 / y0_norm, k_dim)
        y_dt = (
            q_basis
            @ eigen_vectors_t
            @ (np.exp(-1j * scale_factor * eigen_values) * eigen_vectors_t[0, :])
        ) * y0_norm

    else:
        y_dt = [lanczos_expm(A, yi, k_dim, scale_factor) for yi in y0.T]
        y_dt = np.array(y_dt).T

    return y_dt


@requires_array_library("jax")
def jax_lanczos_basis(A: np.ndarray, y0: np.ndarray, k_dim: int):
    """JAX version of lanczos_basis."""

    data_type = jnp.result_type(A.dtype, y0.dtype)
    y0 = y0.astype(data_type)

    projection_0 = A @ y0
    alpha_0 = y0.conj().T @ projection_0
    projection_0 = projection_0 - alpha_0 * y0
    beta_0 = jnp.sqrt(jnp.abs(projection_0.conj().T @ projection_0))

    initial = [y0, projection_0, beta_0]

    def lanczos_iter(carry, _):
        q_p, projection, beta_p = carry

        q_i = projection.T / beta_p
        projection = A @ q_i
        alpha_i = q_i.conj().T @ projection
        projection = projection - alpha_i * q_i - beta_p * q_p
        beta_i = jnp.sqrt(jnp.abs(projection.conj().T @ projection))

        delta = q_i.conj().T @ projection
        projection = projection - delta * q_i
        alpha_i = alpha_i + delta

        carry_next = [q_i, projection, beta_i]
        accumulate = [alpha_i, beta_i, q_i]

        return carry_next, accumulate

    def cond_func(qpb, _):
        _, _, beta_i = qpb

        def zeros_func(_):
            return (
                [jnp.zeros_like(y0), jnp.zeros_like(y0), 0.0],
                [jnp.zeros(1, dtype=data_type)[0], 0.0, jnp.zeros_like(y0)],
            )

        carry_next2, accumulate2 = cond(
            beta_i > 0, lambda carry: lanczos_iter(carry, _), zeros_func, qpb
        )
        return carry_next2, accumulate2

    _, (alpha, beta, q_basis) = scan(cond_func, initial, None, length=k_dim - 1)

    alpha = jnp.array([alpha_0, *alpha])
    beta = jnp.array([beta_0, *beta])
    q_basis = jnp.array([y0, *q_basis])

    tridiagonal = (
        jnp.diag(alpha, k=0) + jnp.diag(beta[: k_dim - 1], k=-1) + jnp.diag(beta[: k_dim - 1], k=1)
    )
    q_basis = q_basis.T
    return tridiagonal, q_basis


@requires_array_library("jax")
def jax_lanczos_eigh(A: np.ndarray, y0: np.ndarray, k_dim: int):
    """JAX version of lanczos_eigh."""

    tridiagonal, q_basis = jax_lanczos_basis(A, y0, k_dim)
    eigen_values, eigen_vectors_t = jnp.linalg.eigh(tridiagonal)

    return q_basis, eigen_values, eigen_vectors_t


@requires_array_library("jax")
def jax_lanczos_expm(
    A: np.ndarray,
    y0: np.ndarray,
    k_dim: int,
    scale_factor: Optional[float] = 1,
):
    """JAX version of lanczos_expm."""

    if y0.ndim == 1:
        A = 1j * A  # make hermitian
        y0_norm = jnp.linalg.norm(y0)
        q_basis, eigen_values, eigen_vectors_t = jax_lanczos_eigh(A, y0 / y0_norm, k_dim)
        y_dt = (
            q_basis
            @ eigen_vectors_t
            @ (jnp.exp(-1j * scale_factor * eigen_values) * eigen_vectors_t[0, :])
        ) * y0_norm

    else:
        _, y_dt = scan(
            lambda _, yi: (None, jax_lanczos_expm(A, yi, k_dim, scale_factor)), None, y0.T
        )
        y_dt = jnp.array(y_dt).T

    return y_dt



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=invalid-name

"""
Utility functions for perturbation module.
"""

from typing import List, Optional, Union
from itertools import product

from multiset import Multiset

from qiskit import QiskitError

from qiskit_dynamics.perturbation.multiset_utils import (
    _validate_non_negative_ints,
    _clean_multisets,
)


def _merge_multiset_expansion_order_labels(
    perturbation_labels: Union[List[int], List[Multiset]],
    expansion_order: Optional[int] = None,
    expansion_labels: Optional[List[Multiset]] = None,
) -> List[Multiset]:
    """Helper function for merging expansion_order and expansion_labels arguments in the multiset
    case for functions that require specifying expansion terms to compute.

    Generates a list of all Multisets of a given size given by expansion_order, and includes any
    additional multisets specified by expansion_labels. The elements of the multisets are drawn from
    perturbation_labels, which is either a list of ints, or a list of Multisets from which the ints
    are drawn.

    At least one of expansion_order or expansion_labels must be specified. Accepts only multisets
    and labels consisting of non-negative integers.

    Args:
        perturbation_labels: Specification of elements of the multisets to generate.
        expansion_order: Size of multisets to generate.
        expansion_labels: Additional multisets to keep.
    Returns:
        List of multisets merging expansion_order and expansion_labels.
    Raises:
        QiskitError: If neither expansion_order nor expansion_labels is specified.
    """

    # validate
    if expansion_order is None and expansion_labels is None:
        raise QiskitError("At least one of expansion_order or expansion_labels must be specified.")

    # clean expansion_labels if specified
    if expansion_labels is not None:
        expansion_labels = _clean_multisets(expansion_labels)
        for label in expansion_labels:
            _validate_non_negative_ints(label)

    # if no expansion_order passed, just return expansion_labels
    if expansion_order is None:
        return expansion_labels

    # generate all multisets of size expansion_order with entries in perturbation_labels

    # get unique multiset elements
    unique_labels = set()
    for perturbation_label in perturbation_labels:
        if isinstance(perturbation_label, int):
            unique_labels = unique_labels.union({perturbation_label})
        else:
            perturbation_label = Multiset(perturbation_label)
            _validate_non_negative_ints(perturbation_label)
            unique_labels = unique_labels.union(perturbation_label.distinct_elements())
    unique_labels = list(unique_labels)
    unique_labels.sort()

    # get all possible counts for multisets of a given size with a given number of labels
    all_multiset_counts = _ordered_partitions(expansion_order, len(unique_labels))

    # create all such multisets
    output_multisets = []
    for multiset_count in all_multiset_counts:
        output_multisets.append(Multiset(dict(zip(unique_labels, multiset_count))))

    if expansion_labels is not None:
        output_multisets = output_multisets + expansion_labels

    return _clean_multisets(output_multisets)


def _merge_list_expansion_order_labels(
    perturbation_num: int,
    expansion_order: Optional[int] = None,
    expansion_labels: Optional[List[List[int]]] = None,
) -> List[int]:
    """Helper function for merging expansion_order and expansion_labels arguments in the list case
    for functions that require specifying expansion terms to compute.

    Generates a list of all lists of integers in [0, ..., perturbation_num - 1] of a given size
    given by expansion_order, and includes any additional lists specified by expansion_labels.

    At least one of expansion_order or expansion_labels must be specified.

    Args:
        perturbation_num: Number of perturbations.
        expansion_order: Size of lists to generate.
        expansion_labels: Additional lists to keep.
    Returns:
        List of multisets merging expansion_order and expansion_labels.
    Raises:
        QiskitError: If neither expansion_order nor expansion_labels is specified.
    """

    # validate
    if expansion_order is None and expansion_labels is None:
        raise QiskitError("At least one of expansion_order or expansion_labels must be specified.")

    if expansion_order is None:
        return expansion_labels

    # generate all possible lists of length expansion_order with integers between
    # [0, ..., perturbation_num - 1]
    unique_indices = list(range(perturbation_num))
    output_lists = list(map(list, product(unique_indices, repeat=expansion_order)))

    if expansion_labels is not None:
        for label in expansion_labels:
            if label not in output_lists:
                output_lists.append(label)

        output_lists.sort(key=str)
        output_lists.sort(key=len)

    return output_lists


def _ordered_partitions(n: int, length: int) -> List[List[int]]:
    """Return the ordered integer partitions of n of a given length, including zeros.

    Args:
        n: Number to partition.
        length: Length of partitions.
    Returns:
        Ordered partitions.
    """

    if length == 1:
        return [[n]]

    full_list = []
    for k in range(n + 1):
        full_list = full_list + [[k] + part for part in _ordered_partitions(n - k, length - 1)]

    return full_list



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=invalid-name

"""
Utility functions for working with multisets.
"""

from typing import List, Optional, Tuple, Iterable
import itertools

from multiset import Multiset

from qiskit import QiskitError


def _validate_non_negative_ints(multiset: Multiset) -> bool:
    """Validate that a multiset only contains non-negative integers."""

    for elem in multiset.distinct_elements():
        if not isinstance(elem, int) or not elem >= 0:
            raise QiskitError(
                "Only Multisets whose entries are non-negative integers are accepted."
            )


def _multiset_to_sorted_list(multiset: Multiset) -> List:
    """Convert multiset to a sorted list. Assumes elements of multiset can be sorted."""

    distinct_elems = list(multiset.distinct_elements())
    distinct_elems.sort()

    sorted_list = []
    for elem in distinct_elems:
        sorted_list = sorted_list + [elem] * multiset[elem]

    return sorted_list


class _MultisetSortKey:
    """Dummy class for usage as a key when sorting Multiset instances. This assumes the elements of
    the multisets can themselves be sorted.
    """

    __slots__ = ("multiset",)

    def __init__(self, multiset: Multiset):
        self.multiset = multiset

    def __lt__(self, other: Multiset) -> bool:
        """Implements an ordering on multisets.

        This orders first according to length (the number of elements in each multiset). If ``self``
        and ``other`` are the same length, ``self < other`` if, when written as fully expanded and
        sorted lists, ``self < other`` in lexicographic ordering. E.g. it holds that
        ``Multiset({0: 2, 1: 1}) < Multiset({0: 1, 1: 2})``, as the list versions are
        ``x = [0, 0, 1]``, and ``y = [0, 1, 1]``. Here ``x[0] == y[0]``, but ``x[1] < y[1]``, and
        hence ``x < y`` in this ordering.
        """
        if len(self.multiset) < len(other.multiset):
            return True

        if len(other.multiset) < len(self.multiset):
            return False

        unique_entries = set(self.multiset.distinct_elements())
        unique_entries.update(other.multiset.distinct_elements())
        unique_entries = sorted(unique_entries)

        for element in unique_entries:
            self_count = self.multiset[element]
            other_count = other.multiset[element]

            if self_count != other_count:
                return self_count > other_count

        return False


def _sorted_multisets(multisets: Iterable[Multiset]) -> List[Multiset]:
    """Sort in non-decreasing order according to the ordering described in the dummy class
    _MultisetSort.
    """
    return sorted(multisets, key=_MultisetSortKey)


def _clean_multisets(multisets: List[Multiset]) -> List[Multiset]:
    """Given a list of multisets, remove duplicates, and sort in non-decreasing order according to
    the _sorted_multisets function.
    """

    unique_multisets = []
    for multiset in multisets:
        multiset = Multiset(multiset)

        if multiset not in unique_multisets:
            unique_multisets.append(multiset)

    return _sorted_multisets(unique_multisets)


def _submultiset_filter(
    multiset_candidates: List[Multiset], multiset_list: List[Multiset]
) -> List[Multiset]:
    """Filter the list of multiset_candidates based on whether they are a submultiset of an element
    in multiset_list.
    """

    filtered_multisets = []
    for candidate in multiset_candidates:
        for multiset in multiset_list:
            if candidate <= multiset:
                filtered_multisets.append(candidate)
                break

    return filtered_multisets


def _submultisets_and_complements(
    multiset: Multiset, submultiset_bound: Optional[int] = None
) -> Tuple[List[Multiset], List[Multiset]]:
    """Return a pair of lists giving all submultisets of size smaller than submultiset_bound, and
    corresponding complements.

    Note: Submultisets and compliments are always strict submultisets.

    Args:
        multiset: The multiset to construct submultisets from.
        submultiset_bound: Strict upper bound on submultiset to include. Defaults to len(multiset).

    Returns:
        Submultisets and corresponding complements.
    """

    if submultiset_bound is None or submultiset_bound > len(multiset):
        submultiset_bound = len(multiset)

    multiset_list = _multiset_to_sorted_list(multiset)

    submultisets = []
    complements = []

    for k in range(1, submultiset_bound):
        location_subsets = itertools.combinations(range(len(multiset_list)), k)
        for location_subset in location_subsets:
            subset = []
            complement = []

            for loc, multiset_entry in enumerate(multiset_list):
                if loc in location_subset:
                    subset.append(multiset_entry)
                else:
                    complement.append(multiset_entry)

            if subset not in submultisets:
                submultisets.append(subset)
                complements.append(complement)

    # convert back to proper dict representation
    formatted_submultisets = [Multiset(submultiset) for submultiset in submultisets]
    formatted_complements = [Multiset(complement) for complement in complements]

    return formatted_submultisets, formatted_complements


def _get_all_submultisets(multisets: List[Multiset]) -> List[Multiset]:
    """Given a list of multisets, return a list of all possible submultisets of multisets in the
    list, including the original multisets.

    This returned list is sorted according to the ordering of the _sorted_multisets function.

    Args:
        multisets: List of multisets (not necessarilly correctly formatted).

    Returns:
        List: Complete list of index multisets generated by the argument.
    """

    if multisets == []:
        return []

    # clean list to unique list of properly formatted terms
    multisets = _clean_multisets(multisets)

    max_order = max(map(len, multisets))

    # partition multisets according to size
    order_dict = {k: [] for k in range(1, max_order + 1)}
    for multiset in multisets:
        order = len(multiset)
        if multiset not in order_dict[order]:
            order_dict[order].append(multiset)

    # loop through orders in reverse order, adding subterms to lower levels if necessary
    for order in range(max_order, 1, -1):
        for multiset in order_dict[order]:
            submultisets = _submultisets_and_complements(multiset, 2)[1]

            for submultiset in submultisets:
                if submultiset not in order_dict[order - 1]:
                    order_dict[order - 1].append(submultiset)

    full_list = []
    for submultisets in order_dict.values():
        full_list += submultisets

    return _sorted_multisets(full_list)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=invalid-name

"""
Magnus expansion-based solver.
"""

from typing import Optional, List, Union

from scipy.linalg import expm
from scipy.integrate._ivp.ivp import OdeResult

from multiset import Multiset

from qiskit.quantum_info import Operator

from qiskit_dynamics import Signal, RotatingFrame, ArrayLike

from .expansion_model import ExpansionModel
from .perturbative_solver import _PerturbativeSolver, _perturbative_solve, _perturbative_solve_jax

try:
    from jax.scipy.linalg import expm as jexpm
except ImportError:
    pass


class MagnusSolver(_PerturbativeSolver):
    """Solver for linear matrix differential equations based on the Magnus expansion.

    This class implements the Magnus expansion-based solver presented in
    [:footcite:`puzzuoli_algorithms_2023`], which is a Magnus expansion variant of the
    *Dysolve* algorithm originally introduced in [:footcite:p:`shillito_fast_2020`]. Its
    setup and behaviour are the same as as the :class:`~qiskit_dynamics.solvers.DysonSolver`
    class, with the sole exception being that it uses a truncated Magnus expansion
    and matrix exponentiation to solve over a single time step. See the
    :ref:`Time-dependent perturbation theory and multi-variable
    series expansions review <perturbation review>` for a description of the Magnus expansion,
    and the documentation for :class:`~qiskit_dynamics.solvers.DysonSolver` for more detailed
    behaviour of this class.
    """

    def __init__(
        self,
        operators: List[Operator],
        rotating_frame: Union[ArrayLike, RotatingFrame, None],
        dt: float,
        carrier_freqs: ArrayLike,
        chebyshev_orders: List[int],
        expansion_order: Optional[int] = None,
        expansion_labels: Optional[List[Multiset]] = None,
        integration_method: Optional[str] = None,
        include_imag: Optional[List[bool]] = None,
        **kwargs,
    ):
        r"""Initialize.

        Args:
            operators: List of constant operators specifying the operators with signal coefficients.
            rotating_frame: Rotating frame to setup the solver in.
                            Must be Hermitian or anti-Hermitian.
            dt: Fixed step size to compile to.
            carrier_freqs: Carrier frequencies of the signals in the generator decomposition.
            chebyshev_orders: Approximation degrees for each signal over the interval [0, dt].
            expansion_order: Order of perturbation terms to compute up to. Specifying this
                             argument results in computation of all terms up to the given order.
                             Can be used in conjunction with ``expansion_terms``.
            expansion_labels: Specific perturbation terms to compute. If both ``expansion_order``
                              and ``expansion_terms`` are specified, then all terms up to
                              ``expansion_order`` are computed, along with the additional terms
                              specified in ``expansion_terms``. Labels are specified either as
                              ``Multiset`` or as valid arguments to the ``Multiset`` constructor.
                              This function further requires that ``Multiset``\s consist only of
                              non-negative integers.
            integration_method: ODE solver method to use when computing perturbation terms.
            include_imag: List of bools determining whether to keep imaginary components in
                          the signal approximation. Defaults to True for all signals.
            kwargs: Additional arguments to pass to the solver when computing perturbation terms.
        """
        model = ExpansionModel(
            operators=operators,
            rotating_frame=rotating_frame,
            dt=dt,
            carrier_freqs=carrier_freqs,
            chebyshev_orders=chebyshev_orders,
            expansion_method="magnus",
            expansion_order=expansion_order,
            expansion_labels=expansion_labels,
            integration_method=integration_method,
            include_imag=include_imag,
            **kwargs,
        )
        super().__init__(model=model)

    def _solve(
        self,
        t0: float,
        n_steps: int,
        y0: ArrayLike,
        signals: List[Signal],
        jax_control_flow: bool = False,
    ) -> OdeResult:
        ys = None
        if jax_control_flow:

            def single_step(x):
                return self.model.Udt @ jexpm(self.model.evaluate(x))

            ys = [y0, _perturbative_solve_jax(single_step, self.model, signals, y0, t0, n_steps)]
        else:

            def single_step(coeffs, y):
                return self.model.Udt @ expm(self.model.evaluate(coeffs)) @ y

            ys = [y0, _perturbative_solve(single_step, self.model, signals, y0, t0, n_steps)]

        return OdeResult(t=[t0, t0 + n_steps * self.model.dt], y=ys)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=invalid-name

"""
Utility functions for Dynamics Backend.
"""


from typing import Optional, Union, List, Dict

import numpy as np
from qiskit import QiskitError
from qiskit.quantum_info import Statevector, DensityMatrix
from qiskit.quantum_info.operators.predicates import is_hermitian_matrix


from qiskit_dynamics.arraylias.alias import ArrayLike, _to_dense
from qiskit_dynamics.models import HamiltonianModel, LindbladModel


def _get_dressed_state_decomposition(
    operator: ArrayLike, rtol=1e-8, atol=1e-5
) -> Union[Dict[str, np.ndarray], List[float], Dict[str, float]]:
    """Get the eigenvalues and eigenvectors of a nearly-diagonal hermitian operator, sorted
    according to overlap with the elementary basis.

    This function is essentially a wrapper around ``numpy.linalg.eigh``, but
    sorts the eigenvectors according to the value of ``np.argmax(np.abs(evec))``. It also
    validates that this is unique for each eigenvector.

    Args:
        operator: Hermitian operator.
        subsystem_dims: Dimensions of the subsystems composing the system.
        rtol: Relative tolerance for Hermiticity check.
        atol: Absolute tolerance for Hermiticity check.

    Returns:
        Tuple: a pair of arrays, one containing eigenvalues and one containing corresponding
        eigenvectors.

    Raises:
        QiskitError: If ``np.argmax(np.abs(evec))`` is non-unique across eigenvectors, or if
        operator is not Hermitian.
    """

    if not is_hermitian_matrix(operator, rtol=rtol, atol=atol):
        raise QiskitError("_get_dressed_state_decomposition received non-Hermitian operator.")

    evals, evecs = np.linalg.eigh(np.array(operator))

    dressed_evals = np.zeros_like(evals)
    dressed_states = np.zeros_like(evecs)

    found_positions = []
    for eigval, evec in zip(evals, evecs.transpose()):
        position = np.argmax(np.abs(evec))
        if position in found_positions:
            raise QiskitError(
                """Dressed-state sorting failed due to non-unique np.argmax(np.abs(evec))
                for eigenvectors."""
            )

        found_positions.append(position)

        dressed_states[:, position] = evec
        dressed_evals[position] = eigval

    return dressed_evals, dressed_states


def _get_lab_frame_static_hamiltonian(model: Union[HamiltonianModel, LindbladModel]) -> np.ndarray:
    """Get the static Hamiltonian in the lab frame and standard basis.

    This function assumes that the model was constructed with operators specified in the lab frame
    (regardless of the rotating frame) and in the standard basis.

    Args:
        model: The model.

    Returns:
        np.ndarray
    """
    static_hamiltonian = None
    if isinstance(model, HamiltonianModel):
        static_hamiltonian = _to_dense(model.static_operator)
    else:
        static_hamiltonian = _to_dense(model.static_hamiltonian)

    static_hamiltonian = 1j * model.rotating_frame.generator_out_of_frame(
        t=0.0, operator=-1j * static_hamiltonian
    )

    return np.array(static_hamiltonian)


def _get_memory_slot_probabilities(
    probability_dict: Dict,
    memory_slot_indices: List[int],
    num_memory_slots: Optional[int] = None,
    max_outcome_value: Optional[int] = None,
) -> Dict:
    """Construct probability dictionary for memory slot outcomes from a probability dictionary for
    state level measurement outcomes.

    Args:
        probability_dict: A list of probabilities for the outcomes of state measurement. Keys
            are assumed to all be strings of integers of the same length.
        memory_slot_indices: Indices of which memory slots store the digits of the keys of
            probability_dict.
        num_memory_slots: Total number of memory slots for results. If None,
            defaults to the maximum index in memory_slot_indices. The default value
            of unused memory slots is 0.
        max_outcome_value: Maximum value that can be stored in a memory slot. All outcomes higher
            than this will be rounded down. Defaults to 1.

    Returns:
        Dict: Keys are memory slot outcomes, values are the probabilities of those outcomes.
    """
    num_memory_slots = num_memory_slots or (max(memory_slot_indices) + 1)
    max_outcome_value = max_outcome_value or 1
    memory_slot_probs = {}
    for level_str, prob in probability_dict.items():
        memory_slot_result = ["0"] * num_memory_slots

        for idx, level in zip(memory_slot_indices, reversed(level_str)):
            if max_outcome_value and int(level) > max_outcome_value:
                level = str(max_outcome_value)
            memory_slot_result[-(idx + 1)] = level

        memory_slot_result = hex(int("".join(memory_slot_result), max_outcome_value + 1))
        if memory_slot_result in memory_slot_probs:
            memory_slot_probs[memory_slot_result] += prob
        else:
            memory_slot_probs[memory_slot_result] = prob

    return memory_slot_probs


def _sample_probability_dict(
    probability_dict: Dict,
    shots: int,
    normalize_probabilities: bool = True,
    seed: Optional[int] = None,
) -> List[str]:
    """Sample outcomes based on probability dictionary.

    Args:
        probability_dict: Dictionary representing probability distribution, with keys being
            outcomes, values being probabilities.
        shots: Number of shots.
        normalize_probabilities: Whether or not to normalize the probabilities to sum to 1 before
            sampling.
        seed: Seed to use in rng construction.

    Return:
        List: of entries of probability_dict, sampled according to the probabilities.
    """
    rng = np.random.default_rng(seed=seed)
    alphabet, probs = zip(*probability_dict.items())

    if normalize_probabilities:
        probs = np.array(probs)
        probs = probs / probs.sum()

    return rng.choice(alphabet, size=shots, replace=True, p=probs)


def _get_counts_from_samples(samples: list) -> Dict:
    """Count items in list."""
    return dict(zip(*np.unique(samples, return_counts=True)))


def _get_subsystem_probabilities(probability_tensor: np.ndarray, sub_idx: int) -> np.ndarray:
    """Marginalize a probability array to a single subsystem. Adapted from
    ``qiskit.quantum_info.QuantumState._subsystem_probabilities``.

    Args:
        probability_tensor: K-dimensional probability array, where the probability of outcome
            ``(idx1, ..., idxk)`` is ``probability_tensor[idx1, ..., idxk]``.
        sub_idx: Subsystem index to return marginalized probabilities.
            ``sub_idx`` is indexed in reverse order to be consistent with qiskit.

    Returns:
        The marginalized probability for the specified subsystem.
    """

    # Convert qargs to tensor axes
    ndim = probability_tensor.ndim
    sub_axis = ndim - 1 - sub_idx

    # Get sum axis for marginalized subsystems
    sum_axis = tuple(i for i in range(ndim) if i != sub_axis)
    if sum_axis:
        probability_tensor = probability_tensor.sum(axis=sum_axis)

    return probability_tensor


def _get_iq_data(
    state: Union[Statevector, DensityMatrix],
    measurement_subsystems: List[int],
    iq_centers: List[List[List[float]]],
    iq_width: float,
    shots: int,
    memory_slot_indices: List[int],
    num_memory_slots: Optional[int] = None,
    seed: Optional[int] = None,
) -> np.ndarray:
    """Generates IQ data for each physical level.

    Args:
        state: Quantum state. measurement_subsystems: Labels of subsystems in the system being
        measured. memory_slot_indices: Indices of which memory slots store the data of subsystems.
        num_memory_slots: Total number of memory slots for results. If None,
            defaults to the maximum index in memory_slot_indices.
        iq_centers: centers for IQ distribution. provided in the format
            ``iq_centers[subsystem][level] = [I,Q]``.
        iq_width: Standard deviation of IQ distribution around the centers. shots: Number of Shots
        seed: Seed for sample generation.

    Returns:
        (I,Q) data as ndarray[shot index, qubit index] = [I,Q]

    Raises:
        QiskitError: If number of centers and levels don't match.
    """
    rng = np.random.default_rng(seed)
    subsystem_dims = [dim for dim in state.dims() if dim != 1]
    probabilities = state.probabilities()
    probabilities_tensor = probabilities.reshape(list(reversed(subsystem_dims)))

    full_i, full_q = [], []
    for sub_idx in measurement_subsystems:
        # Get probabilities for each subsystem
        sub_probability = _get_subsystem_probabilities(probabilities_tensor, sub_idx=sub_idx)
        # No. of shots for each level
        counts_n = rng.multinomial(shots, sub_probability / sum(sub_probability), size=1).T

        if len(counts_n) != len(iq_centers[sub_idx]):
            raise QiskitError(
                f"""Number of centers {len(iq_centers[sub_idx])} not equal
                to number of levels {len(counts_n)}"""
            )

        sub_i, sub_q = [], []
        for idx, count_i in enumerate(counts_n):
            sub_i.append(rng.normal(loc=iq_centers[sub_idx][idx][0], scale=iq_width, size=count_i))
            sub_q.append(rng.normal(loc=iq_centers[sub_idx][idx][1], scale=iq_width, size=count_i))

        full_i.append(np.concatenate(sub_i))
        full_q.append(np.concatenate(sub_q))
    full_iq = np.array([full_i, full_q]).T

    num_memory_slots = num_memory_slots or (max(memory_slot_indices) + 1)
    mem_slot_iq = np.zeros((shots, num_memory_slots, 2))

    for idx, mem_idx in enumerate(memory_slot_indices):
        mem_slot_iq[:, mem_idx, :] = full_iq[:, idx, :]

    return mem_slot_iq



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2021, 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=invalid-name, unidiomatic-typecheck, super-init-not-called

"""
Module for representation of model coefficients.
"""

from typing import List, Callable, Union, Optional, Tuple
import itertools
import operator

import numpy as np
from matplotlib import pyplot as plt

from qiskit import QiskitError
from qiskit_dynamics import ArrayLike
from qiskit_dynamics import DYNAMICS_NUMPY_ALIAS as numpy_alias
from qiskit_dynamics import DYNAMICS_NUMPY as unp
from qiskit_dynamics.arraylias.alias import _numpy_multi_dispatch, _preferred_lib


class Signal:
    r"""General signal class.

    Represents a function of the form:

    .. math::
        Re[f(t)e^{i (2 \pi \nu t + \phi)}]
                = Re[f(t)]\cos(2 \pi \nu t + \phi) - Im[f(t)]\sin(2 \pi \nu t + \phi),

    where

    - :math:`f(t)` is the envelope function.
    - :math:`\nu` is the carrier frequency.
    - :math:`\phi` is the phase.

    The envelope function can be specified either as a constant numeric value (indicating a constant
    function), or as a complex-valued callable, and the frequency and phase must be real.


    .. note::

        :class:`~qiskit_dynamics.signals.Signal` assumes the envelope ``f`` is *array vectorized* in
        the sense that ``f`` can operate on arrays of arbitrary shape and satisfy
        ``f(x)[idx] == f(x[idx])`` for a multidimensional index ``idx``. This can be ensured either
        by writing ``f`` to be vectorized, or by using the ``vectorize`` function in ``numpy`` or
        ``jax.numpy``.

        For example, for an unvectorized envelope function ``f``:

        .. code-block:: python

            import numpy as np
            vectorized_f = np.vectorize(f)
            signal = Signal(envelope=vectorized_f, carrier_freq=2.)
    """

    def __init__(
        self,
        envelope: Union[Callable, ArrayLike],
        carrier_freq: ArrayLike = 0.0,
        phase: ArrayLike = 0.0,
        name: Optional[str] = None,
    ):
        """
        Initializes a signal given by an envelope and a carrier.

        Args:
            envelope: Envelope function of the signal, must be vectorized.
            carrier_freq: Frequency of the carrier. Subclasses such as SignalSums represent the
                carriers of each signal in an array.
            phase: The phase of the carrier. Subclasses such as SignalSums represent the phase of
                each signal in an array.
            name: Name of the signal.
        """
        self._name = name
        self._is_constant = False

        if not callable(envelope):
            # if not callable, we assume a constant
            envelope = unp.asarray(envelope)
            # if envelope is constant and the carrier is zero, this is a constant signal
            try:
                # try block is for catching JAX tracer errors
                if carrier_freq == 0.0:
                    self._is_constant = True
            except Exception:  # pylint: disable=broad-except
                pass

            self._envelope = lambda t: envelope * unp.ones_like(t)
        else:
            self._envelope = envelope

        # set carrier and phase
        self.carrier_freq = carrier_freq
        self.phase = phase

    @property
    def name(self) -> str:
        """Return the name of the signal."""
        return self._name

    @property
    def is_constant(self) -> bool:
        """Whether or not the signal is constant."""
        return self._is_constant

    @property
    def carrier_freq(self) -> ArrayLike:
        """The carrier frequency of the signal."""
        return self._carrier_freq

    @carrier_freq.setter
    def carrier_freq(self, carrier_freq: ArrayLike):
        """Carrier frequency setter. List handling is to support subclasses storing a list of
        frequencies.
        """
        self._carrier_freq = unp.asarray(carrier_freq)
        self._carrier_arg = 1j * 2 * np.pi * self._carrier_freq

    @property
    def phase(self) -> ArrayLike:
        """The phase of the signal."""
        return self._phase

    @phase.setter
    def phase(self, phase: ArrayLike):
        """Phase setter. List handling is to support subclasses storing a list of phases."""
        self._phase = unp.asarray(phase)
        self._phase_arg = 1j * self._phase

    def envelope(self, t: ArrayLike) -> ArrayLike:
        """Vectorized evaluation of the envelope at time t."""
        return self._envelope(t)

    def complex_value(self, t: ArrayLike) -> ArrayLike:
        """Vectorized evaluation of the complex value at time t."""
        arg = self._carrier_arg * t + self._phase_arg
        return self.envelope(t) * unp.exp(arg)

    def __call__(self, t: ArrayLike) -> ArrayLike:
        """Vectorized evaluation of the signal at time(s) t."""
        return unp.real(self.complex_value(t))

    def __str__(self) -> str:
        """Return string representation."""
        if self.name is not None:
            return str(self.name)

        if self.is_constant:
            return f"Constant({str(self(0.0))})"

        return f"Signal(carrier_freq={str(self.carrier_freq)}, phase={str(self.phase)})"

    def __add__(self, other: "Signal") -> "SignalSum":
        return signal_add(self, other)

    def __radd__(self, other: "Signal") -> "SignalSum":
        return self.__add__(other)

    def __mul__(self, other: "Signal") -> "SignalSum":
        return signal_multiply(self, other)

    def __rmul__(self, other: "Signal") -> "SignalSum":
        return self.__mul__(other)

    def __neg__(self) -> "SignalSum":
        return -1 * self

    def __sub__(self, other: "Signal") -> "SignalSum":
        return self + (-other)

    def __rsub__(self, other: "Signal") -> "SignalSum":
        return other + (-self)

    def conjugate(self):
        """Return a new signal whose complex value is the complex conjugate of this one."""

        def conj_env(t):
            return unp.conjugate(self.envelope(t))

        return Signal(conj_env, -self.carrier_freq, -self.phase)

    def draw(
        self,
        t0: float,
        tf: float,
        n: int,
        function: Optional[str] = "signal",
        axis: Optional[plt.axis] = None,
        title: Optional[str] = None,
    ):
        """Plot the signal over an interval.

        The ``function`` arg specifies which function to plot:

            - ``function == 'signal'`` plots the full signal.
            - ``function == 'envelope'`` plots the complex envelope.
            - ``function == 'complex_value'`` plots the ``complex_value``.

        Args:
            t0: Initial time.
            tf: Final time.
            n: Number of points to sample in interval.
            function: Which function to plot.
            axis: The axis to use for plotting.
            title: Title of plot.
        """

        if axis is None:
            plotter = plt
            plotter.title(title)
        else:
            plotter = axis
            plotter.set_title(title)

        t_vals = np.linspace(t0, tf, n)

        y_vals = None
        data_type = "real"
        if function == "signal":
            y_vals = self(t_vals)
            title = title or "Value of " + str(self)
        elif function == "envelope":
            y_vals = self.envelope(t_vals)
            data_type = "complex"
            title = title or "Envelope of " + str(self)
        elif function == "complex_value":
            y_vals = self.complex_value(t_vals)
            data_type = "complex"
            title = title or "Complex value of " + str(self)

        legend = False
        if data_type == "complex":
            plotter.plot(t_vals, np.real(y_vals), label="Real")
            plotter.plot(t_vals, np.imag(y_vals), label="Imag")
            legend = True
        else:
            plotter.plot(t_vals, y_vals)

        if legend:
            plotter.legend()


class DiscreteSignal(Signal):
    r"""Piecewise constant signal implemented as an array of samples.

    The envelope is specified by an array of samples ``s = [s_0, ..., s_k]``, sample width ``dt``,
    and a start time ``t_0``, with the envelope being evaluated as
    :math:`f(t) =` ``s[floor((t - t0)/dt)]`` if ``t`` is in the interval with endpoints
    ``start_time`` and ``start_time + dt * len(samples)``, and ``0.0`` otherwise. By default a
    :class:`~qiskit_dynamics.signals.DiscreteSignal` is defined to start at :math:`t=0` but a custom
    start time can be set via the ``start_time`` kwarg.
    """

    def __init__(
        self,
        dt: float,
        samples: ArrayLike,
        start_time: float = 0.0,
        carrier_freq: ArrayLike = 0.0,
        phase: ArrayLike = 0.0,
        name: str = None,
    ):
        """Initialize a piecewise constant signal.

        Args:
            dt: The duration of each sample.
            samples: The array of samples.
            start_time: The time at which the signal starts.
            carrier_freq: Frequency of the carrier. Subclasses such as SignalSums represent the
                carriers of each signal in an array.
            phase: The phase of the carrier. Subclasses such as SignalSums represent the phase of
                each signal in an array.
            name: Name of the signal.
        """
        self._dt = dt

        samples = unp.asarray(samples)

        if len(samples) == 0:
            zero_pad = np.asarray([0])
        else:
            zero_pad = unp.expand_dims(unp.zeros_like(samples[0]), 0)
        self._padded_samples = unp.append(samples, zero_pad, axis=0)

        self._start_time = start_time

        # define internal envelope function
        def envelope(t):
            t = unp.asarray(t)
            idx = unp.clip(
                unp.array((t - self._start_time) // self._dt, dtype=int),
                -1,
                len(self.samples),
            )
            return numpy_alias(like=_preferred_lib(self._padded_samples, idx)).asarray(
                self._padded_samples
            )[idx]

        Signal.__init__(self, envelope=envelope, carrier_freq=carrier_freq, phase=phase, name=name)

    @classmethod
    def from_Signal(
        cls,
        signal: Signal,
        dt: float,
        n_samples: int,
        start_time: Optional[float] = 0.0,
        sample_carrier: Optional[bool] = False,
    ):
        r"""Constructs a ``DiscreteSignal`` object by sampling a ``Signal``\.

        The optional argument ``sample_carrier`` controls whether or not to include the carrier in
        the sampling. I.e.:

            - If ``sample_carrier == False``\, a ``DiscreteSignal`` is constructed with:
                - ``samples`` obtained by sampling ``signal.envelope``\.
                - ``carrier_freq = signal.carrier_freq``\.
                - ``phase = signal.phase``\.

            - If ``sample_carrier == True``\, a ``DiscreteSignal`` is constructed with:
                - ``samples`` obtained by sampling ``signal`` (as a ``callable``\)
                - ``carrier_freq = 0``\.
                - ``phase = signal.phase``\.

        In either case, samples are obtained from the midpoint of each interval.

        Args:
            signal: Signal to sample.
            dt: Time increment to use.
            n_samples: Number of steps to resample with.
            start_time: Start time from which to resample.
            sample_carrier: Whether or not to include the carrier in the sampling.

        Returns:
            DiscreteSignal: The discretized ``Signal``\.
        """

        times = start_time + (np.arange(n_samples) + 0.5) * dt
        freq = signal.carrier_freq

        if sample_carrier:
            freq = 0.0
            samples = signal(times)
        else:
            samples = signal.envelope(times)

        return DiscreteSignal(
            dt,
            samples,
            start_time=start_time,
            carrier_freq=freq,
            phase=signal.phase,
            name=signal.name,
        )

    @property
    def duration(self) -> int:
        """
        Returns:
            duration: The duration of the signal in samples.
        """
        return len(self.samples)

    @property
    def dt(self) -> float:
        """
        Returns:
             dt: the duration of each sample.
        """
        return self._dt

    @property
    def samples(self) -> ArrayLike:
        """
        Returns:
            samples: the samples of the piecewise constant signal.
        """
        return self._padded_samples[:-1]

    @property
    def start_time(self) -> float:
        """
        Returns:
            start_time: The time at which the list of samples start.
        """
        return self._start_time

    def conjugate(self):
        return self.__class__(
            dt=self._dt,
            samples=unp.conjugate(self.samples),
            start_time=self._start_time,
            carrier_freq=-self.carrier_freq,
            phase=-self.phase,
        )

    def add_samples(self, start_sample: int, samples: List):
        """
        Appends samples to the pulse starting at start_sample, filling any gap with zeros.

        Args:
            start_sample: Index of the sample at which the new samples should be appended.
            samples: List of samples to append.

        Raises:
            QiskitError: If start_sample is less than the current length of samples.
        """
        samples = unp.asarray(samples)

        if len(samples) < 1:
            return

        if start_sample < len(self.samples):
            raise QiskitError("Samples can only be added afer the last sample.")

        zero_pad = unp.expand_dims(unp.zeros_like(samples[0]), 0)

        new_samples = self.samples
        if len(self.samples) < start_sample:
            new_samples = unp.append(
                new_samples, unp.repeat(zero_pad, start_sample - len(self.samples))
            )

        new_samples = _numpy_multi_dispatch(new_samples, samples, path="append")
        self._padded_samples = unp.append(new_samples, zero_pad, axis=0)

    def __str__(self) -> str:
        """Return string representation."""
        if self.name is not None:
            return str(self.name)

        return f"""DiscreteSignal(dt={self.dt}, carrier_freq={str(self.carrier_freq)},
                phase={str(self.phase)})"""


class SignalCollection:
    """Base class for a list-like collection of signals."""

    def __init__(self, signal_list: List[Signal]):
        """Initialize by storing list of signals.

        Args:
            signal_list: List of signals.
        """
        self._is_constant = False
        self._components = signal_list

    @property
    def components(self) -> List[Signal]:
        """The list of components."""
        return self._components

    def __len__(self):
        """Number of components."""
        return len(self.components)

    def __getitem__(self, idx: Union[ArrayLike, slice]) -> Union[Signal, "SignalCollection"]:
        """Get item with NumPy-style subscripting, as if this class were a 1d array."""

        if type(idx) != slice and unp.asarray(idx).ndim > 0:
            idx = list(idx)

        # get a list of the subcomponents
        if type(idx) == list:
            # handle lists
            sublist = operator.itemgetter(*idx)(self.components)

            # output will either be a single signal or a tuple of Signals
            # convert to list if tuple
            if type(sublist) == tuple:
                sublist = list(sublist)
        else:
            # handle slices or singletons
            sublist = operator.itemgetter(idx)(self.components)

        # at this point sublist should either be a single Signal, or a list of Signals
        if isinstance(sublist, list):
            return self.__class__(sublist)
        else:
            return sublist

    def __iter__(self):
        """Return iterator over component list."""
        return self.components.__iter__()

    def conjugate(self) -> "SignalCollection":
        """Return the conjugation of this collection."""
        return self.__class__([sig.conjugate() for sig in self.components])


class SignalSum(SignalCollection, Signal):
    r"""Represents a sum of signals.

    I.e. a sum of ``Signal`` objects:

    .. math::
        s_1(t) + \dots + s_k(t)

    For basic evaluation, this class behaves in the same manner as ``Signal``:

        - ``__call__`` evaluates the sum.
        - ``complex_value`` evaluates the sum of the complex values of the individual summands.

    Attributes ``carrier_freq`` and ``phase`` here correspond to an ``ArrayLike`` of
    frequencies/phases for each term in the sum, and the ``envelope`` method returns an
    ``ArrayLike`` of the envelopes for each summand.

    Internally, the signals are stored as a list in the ``components`` attribute, which can be
    accessed via direct subscripting of the object.
    """

    def __init__(self, *signals, name: Optional[str] = None):
        r"""Initialize with a list of Signal objects through ``args``\.

        Args:
            signals: ``Signal`` subclass objects.
            name: Name of the sum.

        Raises:
            QiskitError: If ``signals`` are not subclasses of ``Signal``\.
        """
        self._name = name

        components = []
        for sig in signals:
            if isinstance(sig, list):
                sig = SignalSum(*sig)

            if isinstance(sig, SignalSum):
                components += sig.components
            elif isinstance(sig, Signal):
                components.append(sig)
            else:
                try:
                    if unp.asarray(sig).ndim == 0:
                        components.append(Signal(sig))
                except QiskitError as qe:
                    raise QiskitError(
                        "Components of a SignalSum must be instances "
                        "of a Signal subclass or a scalar."
                    ) from qe

        SignalCollection.__init__(self, components)

        def envelope(t):
            return unp.moveaxis(unp.asarray([sig.envelope(t) for sig in self.components]), 0, -1)

        carrier_freqs = []
        for sig in self.components:
            carrier_freqs.append(sig.carrier_freq)

        phases = []
        for sig in self.components:
            phases.append(sig.phase)

        Signal.__init__(
            self, envelope=envelope, carrier_freq=carrier_freqs, phase=phases, name=name
        )

    def complex_value(self, t: ArrayLike) -> ArrayLike:
        """Return the sum of the complex values of each component."""
        exp_phases = unp.exp(unp.expand_dims(t, -1) * self._carrier_arg + self._phase_arg)
        return unp.sum(self.envelope(t) * exp_phases, axis=-1)

    def __str__(self):
        if self.name is not None:
            return str(self.name)

        if len(self) == 0:
            return "SignalSum()"

        default_str = str(self[0])
        for sig in self.components[1:]:
            default_str += f" + {str(sig)}"

        return default_str

    def flatten(self) -> Signal:
        r"""Merge into a single ``Signal``\. The output frequency is given by the
        average.
        """

        if len(self) == 0:
            return Signal(0.0)
        elif len(self) == 1:
            return self.components[0]

        ave_freq = unp.sum(self.carrier_freq) / len(self)
        shifted_arg = self._carrier_arg - (1j * 2 * np.pi * ave_freq)

        def merged_env(t):
            exp_phases = unp.exp(unp.expand_dims(t, -1) * shifted_arg + self._phase_arg)
            return unp.sum(self.envelope(t) * exp_phases, axis=-1)

        return Signal(envelope=merged_env, carrier_freq=ave_freq, name=str(self))


class DiscreteSignalSum(DiscreteSignal, SignalSum):
    """Represents a sum of piecewise constant signals, all with the same time parameters: dt, number
    of samples, and start time.
    """

    def __init__(
        self,
        dt: float,
        samples: ArrayLike,
        start_time: float = 0.0,
        carrier_freq: ArrayLike = None,
        phase: ArrayLike = None,
        name: str = None,
    ):
        r"""Directly initialize a ``DiscreteSignalSum``\. Samples of all terms in the
        sum are specified as a 2d array, with 0th axis indicating time, and 1st axis indicating a
        term in the sum.

        Args:
            dt: The duration of each sample.
            samples: The 2d array representing a list whose elements are all envelope values at a
                given time.
            start_time: The time at which the signal starts.
            carrier_freq: Array with the carrier frequencies of each term in the sum.
            phase: Array with the phases of each term in the sum.
            name: name of the signal.
        """

        samples = unp.asarray(samples)
        if carrier_freq is None:
            carrier_freq = np.zeros(samples.shape[-1], dtype=float)

        if phase is None:
            phase = np.zeros(samples.shape[-1], dtype=float)

        DiscreteSignal.__init__(
            self,
            dt=dt,
            samples=samples,
            start_time=start_time,
            carrier_freq=carrier_freq,
            phase=phase,
            name=name,
        )

        # construct individual components so they can be accessed as in SignalSum
        components = []
        for sample_row, freq, phi in zip(self.samples.transpose(), carrier_freq, phase):
            components.append(
                DiscreteSignal(
                    dt=self.dt,
                    samples=sample_row,
                    start_time=self.start_time,
                    carrier_freq=freq,
                    phase=phi,
                )
            )

        self._components = components

    @classmethod
    def from_SignalSum(
        cls,
        signal_sum: SignalSum,
        dt: float,
        n_samples: int,
        start_time: Optional[float] = 0.0,
        sample_carrier: Optional[bool] = False,
    ):
        r"""Constructs a ``DiscreteSignalSum`` object by sampling a ``SignalSum``\.

        The optional argument ``sample_carrier`` controls whether or not to include the carrier in
        the sampling. I.e.:

            - If ``sample_carrier == False``, a ``DiscreteSignalSum`` is constructed with:
                - ``samples`` obtained by sampling ``signal_sum.envelope``\.
                - ``carrier_freq = signal_sum.carrier_freq``\.
                - ``phase = signal_sum.phase``\.

            - If ``sample_carrier == True``\, a ``DiscreteSignal`` is constructed with:
                - ``samples`` obtained by sampling ``signal_sum`` (as a ``callable``\)
                - ``carrier_freq = 0``\.
                - ``phase = signal_sum.phase``\.

        In either case, samples are obtained from the midpoint of each interval.

        Args:
            signal_sum: SignalSum to sample.
            dt: Time increment to use.
            n_samples: Number of steps to resample with.
            start_time: Start time from which to resample.
            sample_carrier: Whether or not to include the carrier in the sampling.

        Returns:
            DiscreteSignalSum: The discretized ``SignalSum``\.
        """

        times = start_time + (np.arange(n_samples) + 0.5) * dt

        freq = signal_sum.carrier_freq

        if sample_carrier:
            freq = 0.0 * freq
            exp_phases = unp.exp(unp.expand_dims(times, -1) * signal_sum._carrier_arg)
            samples = signal_sum.envelope(times) * exp_phases
        else:
            samples = signal_sum.envelope(times)

        return DiscreteSignalSum(
            dt,
            samples,
            start_time=start_time,
            carrier_freq=freq,
            phase=signal_sum.phase,
            name=signal_sum.name,
        )

    def __str__(self):
        """Get the string rep."""
        if self.name is not None:
            return str(self.name)

        if len(self) == 0:
            return "DiscreteSignalSum()"

        default_str = str(self[0])
        for sig in self.components[1:]:
            default_str += f" + {str(sig)}"

        return default_str

    def __getitem__(self, idx: ArrayLike) -> Signal:
        """Enables numpy-style subscripting, as if this class were a 1d array."""

        if type(idx) == int and idx >= len(self):
            raise IndexError("index out of range for DiscreteSignalSum of length " + str(len(self)))

        samples = self.samples[:, idx]
        carrier_freqs = self.carrier_freq[idx]
        phases = self.phase[idx]

        if samples.ndim == 1:
            samples = unp.asarray([samples])

        if carrier_freqs.ndim == 0:
            carrier_freqs = unp.asarray([carrier_freqs])

        if phases.ndim == 0:
            phases = unp.asarray([phases])

        if len(samples) == 1:
            return DiscreteSignal(
                dt=self.dt,
                samples=samples[0],
                start_time=self.start_time,
                carrier_freq=carrier_freqs[0],
                phase=phases[0],
            )

        return DiscreteSignalSum(
            dt=self.dt,
            samples=samples,
            start_time=self.start_time,
            carrier_freq=carrier_freqs,
            phase=phases,
        )


class SignalList(SignalCollection):
    r"""A list of signals with functionality for simultaneous evaluation.

    The passed list is stored in the ``components`` attribute.
    """

    def __init__(self, signal_list: List[Signal]):
        signal_list = [to_SignalSum(signal) for signal in signal_list]

        super().__init__(signal_list)

        # setup complex value and full signal evaluation
        self._eval_complex_value = lambda t: unp.asarray(
            [sig.complex_value(t) for sig in self.components]
        )
        self._eval_signals = lambda t: unp.asarray([sig(t) for sig in self.components])

    def complex_value(self, t: ArrayLike) -> ArrayLike:
        """Vectorized evaluation of complex value of components."""
        return unp.moveaxis(self._eval_complex_value(t), 0, -1)

    def __call__(self, t: ArrayLike) -> ArrayLike:
        """Vectorized evaluation of all components."""
        return unp.moveaxis(self._eval_signals(t), 0, -1)

    def flatten(self) -> "SignalList":
        """Return a ``SignalList`` with each component flattened."""
        flattened_list = []
        for sig in self.components:
            if isinstance(sig, SignalSum):
                flattened_list.append(sig.flatten())
            else:
                flattened_list.append(sig)

        return SignalList(flattened_list)

    @property
    def drift(self) -> ArrayLike:
        r"""Return the drift ``ArrayLike``\, i.e. return an ``ArrayLike`` whose entries are the sum
        of the constant parts of the corresponding component of this ``SignalList``\.
        """

        drift_array = []
        for sig_entry in self.components:
            val = 0.0

            if not isinstance(sig_entry, SignalSum):
                sig_entry = SignalSum(sig_entry)

            for term in sig_entry:
                if term.is_constant:
                    val += term(0.0)

            drift_array.append(val)

        return unp.asarray(drift_array)


def signal_add(sig1: Signal, sig2: Signal) -> SignalSum:
    """Add two signals."""

    # generic routine
    # convert to SignalSum instances
    try:
        sig1 = to_SignalSum(sig1)
        sig2 = to_SignalSum(sig2)
    except QiskitError as qe:
        raise QiskitError("Only a number or a Signal instance can be added to a Signal.") from qe

    # if both are DiscreteSignalSum objects with compatible structure, append data together
    if isinstance(sig1, DiscreteSignalSum) and isinstance(sig2, DiscreteSignalSum):
        if (
            sig1.dt == sig2.dt
            and sig1.start_time == sig2.start_time
            and sig1.duration == sig2.duration
        ):
            samples = _numpy_multi_dispatch(sig1.samples, sig2.samples, axis=1, path="append")
            carrier_freq = _numpy_multi_dispatch(
                sig1.carrier_freq, sig2.carrier_freq, path="append"
            )
            phase = _numpy_multi_dispatch(sig1.phase, sig2.phase, path="append")
            return DiscreteSignalSum(
                dt=sig1.dt,
                samples=samples,
                start_time=sig1.start_time,
                carrier_freq=carrier_freq,
                phase=phase,
            )

    sig_sum = SignalSum(*(sig1.components + sig2.components))

    return sig_sum


def signal_multiply(sig1: Signal, sig2: Signal) -> SignalSum:
    r"""Multiply two ``Signal``\s. For a pair of elementary (non-``SignalSum``\) ``Signal``\s,
    expands the product of two signals into a ``SignalSum`` via the formula:

    .. math::
        Re[f(t)e^{i(2 \pi \nu t + \phi)}] \times Re[g(t)e^{i(2 \pi \omega t + \psi)}]
         = Re[\frac{1}{2} f(t)g(t)e^{i(2\pi (\omega + \nu)t + (\phi + \psi))} ]
          + Re[\frac{1}{2} f(t)\overline{g(t)}e^{i(2\pi (\omega - \nu)t + (\phi - \psi))} ]

    If either (or both) of ``sig1`` or ``sig2`` are ``SignalSum``\s, the multiplication is
    distributed over addition.
    """

    # convert to SignalSum instances
    try:
        sig1 = to_SignalSum(sig1)
        sig2 = to_SignalSum(sig2)
    except QiskitError as qe:
        raise QiskitError("Only a number or a Signal instance can multiply a Signal.") from qe

    sig1, sig2 = sort_signals(sig1, sig2)

    # if sig1 contains only a constant and sig2 is a DiscreteSignalSum
    if len(sig1) == 1 and sig1[0].is_constant and isinstance(sig2, DiscreteSignalSum):
        return DiscreteSignalSum(
            dt=sig2.dt,
            samples=sig1(0.0) * sig2.samples,
            start_time=sig2.start_time,
            carrier_freq=sig2.carrier_freq,
            phase=sig2.phase,
        )
    # if both are DiscreteSignalSum objects with compatible structure, append data together
    elif isinstance(sig1, DiscreteSignalSum) and isinstance(sig2, DiscreteSignalSum):
        if (
            sig1.dt == sig2.dt
            and sig1.start_time == sig2.start_time
            and sig1.duration == sig2.duration
        ):
            # this vectorized operation produces a 2d array whose columns are the products of
            # the original columns
            new_samples = unp.asarray(
                0.5
                * (sig1.samples[:, :, None] * sig2.samples[:, None, :]).reshape(
                    (sig1.samples.shape[0], sig1.samples.shape[1] * sig2.samples.shape[1]),
                    order="C",
                )
            )
            new_samples_conj = unp.asarray(
                0.5
                * (sig1.samples[:, :, None] * sig2.samples[:, None, :].conj()).reshape(
                    (sig1.samples.shape[0], sig1.samples.shape[1] * sig2.samples.shape[1]),
                    order="C",
                )
            )
            samples = _numpy_multi_dispatch(new_samples, new_samples_conj, axis=1, path="append")

            new_freqs = sig1.carrier_freq + sig2.carrier_freq
            new_freqs_conj = sig1.carrier_freq - sig2.carrier_freq
            freqs = _numpy_multi_dispatch(
                unp.asarray(new_freqs), unp.asarray(new_freqs_conj), path="append"
            )

            new_phases = sig1.phase + sig2.phase
            new_phases_conj = sig1.phase - sig2.phase
            phases = _numpy_multi_dispatch(
                unp.asarray(new_phases), unp.asarray(new_phases_conj), path="append"
            )

            return DiscreteSignalSum(
                dt=sig1.dt,
                samples=samples,
                start_time=sig1.start_time,
                carrier_freq=freqs,
                phase=phases,
            )

    # initialize to empty sum
    product = SignalSum()

    # loop through every pair of components and multiply
    for comp1, comp2 in itertools.product(sig1.components, sig2.components):
        product += base_signal_multiply(comp1, comp2)

    return product


def base_signal_multiply(sig1: Signal, sig2: Signal) -> Signal:
    r"""Utility function for multiplying two elementary (non ``SignalSum``\) signals. This function
    assumes ``sig1`` and ``sig2`` are legitimate instances of ``Signal`` subclasses.

    Special cases:

        - Multiplication of two constant ``Signal``\s returns a constant ``Signal``\.
        - Multiplication of a constant ``Signal`` and a ``DiscreteSignal`` returns a
          ``DiscreteSignal``\.
        - If two ``DiscreteSignal``\s have compatible parameters, the resulting signals are
          ``DiscreteSignal``\, with the multiplication being implemented by array multiplication of
          the samples.
        - Lastly, if no special rules apply, the two ``Signal``\s are multiplied generically via
        multiplication of the envelopes as functions.

    When a sum with two signals is produced, the carrier frequency (phase) of each component are,
    respectively, the sum and difference of the two frequencies (phases). For special cases
    involving constant ``Signal``\s and a non-constant ``Signal``, the carrier frequency and phase
    are preserved as that of the non-constant piece.

    Args:
        sig1: First signal.
        sig2: Second signal.

    Returns:
        SignalSum: Representing the RHS of the formula when two Signals are multiplied.
    """

    # ensure signals are ordered from most to least specialized
    sig1, sig2 = sort_signals(sig1, sig2)

    if sig1.is_constant and sig2.is_constant:
        return Signal(sig1(0.0) * sig2(0.0))
    elif sig1.is_constant and type(sig2) is DiscreteSignal:
        # multiply the samples by the constant
        return DiscreteSignal(
            dt=sig2.dt,
            samples=sig1(0.0) * sig2.samples,
            start_time=sig2.start_time,
            carrier_freq=sig2.carrier_freq,
            phase=sig2.phase,
        )
    elif sig1.is_constant and type(sig2) is Signal:
        const = sig1(0.0)

        def new_env(t):
            return const * sig2.envelope(t)

        return Signal(envelope=new_env, carrier_freq=sig2.carrier_freq, phase=sig2.phase)
    elif type(sig1) is DiscreteSignal and type(sig2) is DiscreteSignal:
        # verify compatible parameters before applying special rule
        if (
            sig1.start_time == sig2.start_time
            and sig1.dt == sig2.dt
            and len(sig1.samples) == len(sig2.samples)
        ):
            pwc1 = DiscreteSignal(
                dt=sig2.dt,
                samples=0.5 * sig1.samples * sig2.samples,
                start_time=sig2.start_time,
                carrier_freq=sig1.carrier_freq + sig2.carrier_freq,
                phase=sig1.phase + sig2.phase,
            )
            pwc2 = DiscreteSignal(
                dt=sig2.dt,
                samples=0.5 * sig1.samples * unp.conjugate(sig2.samples),
                start_time=sig2.start_time,
                carrier_freq=sig1.carrier_freq - sig2.carrier_freq,
                phase=sig1.phase - sig2.phase,
            )
            return pwc1 + pwc2

    # if no special cases apply, implement generic rule
    def new_env1(t):
        return 0.5 * sig1.envelope(t) * sig2.envelope(t)

    def new_env2(t):
        return 0.5 * sig1.envelope(t) * unp.conjugate(sig2.envelope(t))

    prod1 = Signal(
        envelope=new_env1,
        carrier_freq=sig1.carrier_freq + sig2.carrier_freq,
        phase=sig1.phase + sig2.phase,
    )
    prod2 = Signal(
        envelope=new_env2,
        carrier_freq=sig1.carrier_freq - sig2.carrier_freq,
        phase=sig1.phase - sig2.phase,
    )
    return prod1 + prod2


def sort_signals(sig1: Signal, sig2: Signal) -> Tuple[Signal, Signal]:
    r"""Utility function for ordering a pair of ``Signal``\s according to the partial order:
    ``sig1 <= sig2`` if and only if:
        - ``type(sig1)`` precedes ``type(sig2)`` in the list
        ``[constant, DiscreteSignal, Signal, SignalSum, DiscreteSignalSum]``\.
    """
    if sig1.is_constant:
        pass
    elif sig2.is_constant:
        sig1, sig2 = sig2, sig1
    elif isinstance(sig1, DiscreteSignal) and not isinstance(sig1, DiscreteSignalSum):
        pass
    elif isinstance(sig2, DiscreteSignal) and not isinstance(sig2, DiscreteSignalSum):
        sig2, sig1 = sig1, sig2
    elif isinstance(sig1, Signal) and not isinstance(sig1, SignalSum):
        pass
    elif isinstance(sig2, Signal) and not isinstance(sig2, SignalSum):
        sig2, sig1 = sig1, sig2
    elif isinstance(sig1, SignalSum) and not isinstance(sig1, DiscreteSignalSum):
        pass
    elif isinstance(sig2, SignalSum) and not isinstance(sig2, DiscreteSignalSum):
        sig2, sig1 = sig1, sig2
    elif isinstance(sig1, DiscreteSignalSum):
        pass
    elif isinstance(sig2, DiscreteSignalSum):
        sig2, sig1 = sig1, sig2

    return sig1, sig2


def to_SignalSum(sig: Union[ArrayLike, Signal]) -> SignalSum:
    r"""Convert the input to a SignalSum according to:

        - If it is already a ``SignalSum``\, do nothing.
        - If it is a Signal but not a ``SignalSum``\, wrap in a ``SignalSum``\.
        - If it is a number, wrap in constant ``Signal`` in a ``SignalSum``\.
        - Otherwise, raise an error.

    Args:
        sig: A SignalSum compatible input.

    Returns:
        SignalSum

    Raises:
        QiskitError: If the input type is incompatible with SignalSum.
    """

    if isinstance(sig, (int, float, complex)) or (
        not isinstance(sig, (int, float, complex, list, Signal)) and sig.ndim == 0
    ):
        return SignalSum(Signal(sig))
    elif isinstance(sig, DiscreteSignal) and not isinstance(sig, DiscreteSignalSum):
        if sig.samples.shape == (0,):
            new_samples = unp.asarray([sig.samples])
        else:
            new_samples = unp.asarray([sig.samples]).transpose(1, 0)
        return DiscreteSignalSum(
            dt=sig.dt,
            samples=new_samples,
            start_time=sig.start_time,
            carrier_freq=unp.asarray([sig.carrier_freq]),
            phase=unp.asarray([sig.phase]),
        )
    elif isinstance(sig, Signal) and not isinstance(sig, SignalSum):
        return SignalSum(sig)
    elif isinstance(sig, SignalSum):
        return sig

    raise QiskitError("Input type incompatible with SignalSum.")



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Registering conjugate.
"""


def register_conjugate(alias):
    """Register linear functions for each array library."""

    try:
        from jax.dtypes import canonicalize_dtype
        import jax.numpy as jnp
        from jax.experimental.sparse import sparsify

        # can be changed to sparsify(jnp.conjugate) when implemented
        def conj_workaround(x):
            if jnp.issubdtype(x.dtype, canonicalize_dtype(jnp.complex128)):
                return x.real - 1j * x.imag
            return x

        alias.register_function(func=sparsify(conj_workaround), lib="jax_sparse", path="conjugate")

    except ImportError:
        pass



# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=invalid-name

"""
Pulse schedule to Signals converter.
"""

from typing import Callable, Dict, List, Optional
import functools
from warnings import warn

import numpy as np
import sympy as sym

from qiskit.pulse import (
    Schedule,
    Play,
    ShiftPhase,
    SetPhase,
    ShiftFrequency,
    SetFrequency,
    Waveform,
    MeasureChannel,
    DriveChannel,
    ControlChannel,
    AcquireChannel,
)
from qiskit.pulse.exceptions import PulseError
from qiskit.pulse.library import SymbolicPulse
from qiskit import QiskitError

from qiskit_dynamics import DYNAMICS_NUMPY as unp
from qiskit_dynamics import ArrayLike

from qiskit_dynamics.signals import DiscreteSignal

try:
    import jax
    import jax.numpy as jnp
except ImportError:
    pass


class InstructionToSignals:
    """Converts pulse instructions to signals to be used in models.

    The :class:`InstructionsToSignals` class converts a pulse schedule to a list of signals that can
    be given to a model. This conversion is done by calling the :meth:`get_signals` method on a
    schedule. The converter applies to instances of :class:`~qiskit.pulse.Schedule`. Instances of
    :class:`~qiskit.pulse.ScheduleBlock` must first be converted to :class:`~qiskit.pulse.Schedule`
    using the :func:`~qiskit.pulse.transforms.block_to_schedule` function in Qiskit Pulse.

    The converter can be initialized with the optional arguments ``carriers`` and ``channels``. When
    ``channels`` is given, only the signals specified by name in ``channels`` are returned. The
    ``carriers`` dictionary specifies the analog carrier frequency of each channel. Here, the keys
    are the channel name, e.g. ``d12`` for drive channel number ``12``, and the values are the
    corresponding frequency. If a channel is not present in ``carriers`` it is assumed that the
    analog carrier frequency is zero.

    See the :meth:`get_signals` method documentation for a detailed description of how pulse
    schedules are interpreted and translated into :class:`.DiscreteSignal` objects.
    """

    def __init__(
        self,
        dt: float,
        carriers: Optional[Dict[str, float]] = None,
        channels: Optional[List[str]] = None,
    ):
        """Initialize pulse schedule to signals converter.

        Args:
            dt: Length of the samples. This is required by the converter as pulse schedule are
                specified in units of dt and typically do not carry the value of dt with them.
            carriers: A dict of analog carrier frequencies. The keys are the names of the channels
                and the values are the corresponding carrier frequency.
            channels: A list of channels that the :meth:`get_signals` method should return. This
                argument will cause :meth:`get_signals` to return the signals in the same order as
                the channels. Channels present in the schedule but absent from channels will not be
                included in the returned object. If None is given (the default) then all channels
                present in the pulse schedule are returned.
        """

        self._dt = dt
        self._channels = channels
        self._carriers = carriers or {}

    def get_signals(self, schedule: Schedule) -> List[DiscreteSignal]:
        r"""Convert a schedule to a corresponding list of DiscreteSignal instances.

        Which channels are converted, and the order they are returned, is controlled by the
        ``channels`` argument at instantiation. The ``carriers`` instantiation argument sets the
        analog carrier frequency for each channel, which is fixed for the full duration. For a given
        channel, the :math:`k^{th}` envelope sample for the corresponding :class:`.DiscreteSignal`
        is determined according to the following formula:

        .. math::
            f(k) \exp(i(2\pi \Delta\nu(k) k dt + \phi(k) + 2 \pi \phi_a(k))),

        where:

        * :math:`f(k)` is the waveform value at the :math:`k^{th}` time step as specified by
          ``Play`` instructions.
        * :math:`\Delta\nu(k)` is the frequency deviation at time step :math:`k` from the analog
          carrier as the result of ``SetFrequency`` and ``ShiftFrequency`` instructions. As evident
          by the formula, carrier frequency deviations as a result of these instructions are handled
          digitally, with the analog carrier frequency being fixed for the entirety of the schedule.
        * :math:`dt` is the sample rate as specified by the ``dt`` instantiation argument.
        * :math:`\phi(k)` is the channel phase at time step :math:`k`, as determined by
          ``ShiftPhase`` and ``SetPhase`` instructions.
        * :math:`\phi_a(k)` is the phase correction term at time step :math:`k`, impacted by
          ``SetFrequency`` and ``ShiftFrequency`` instructions, described below.

        In detail, the sample array for the output signal for each channel is generated by iterating
        over each instruction in the schedule in temporal order. New samples are appended with every
        ``Play`` instruction on the given channel, using the waveform values and the current value
        of the tracked parameters :math:`\Delta\nu`, :math:`\phi`, and :math:`\phi_a`, which are
        initialized to :math:`0`. Explicitly, each instruction is interpreted as follows:

        * ``Play`` instructions add new samples to the sample array, according to the above formula,
          using the waveform specified in the instruction and the current values of
          :math:`\Delta\nu`, :math:`\phi`, and :math:`\phi_a`.
        * ``ShiftPhase``, with a phase value :math:`\psi`, updates :math:`\phi \mapsto \phi + \psi`.
        * ``SetPhase``, with a phase value :math:`\psi`, updates :math:`\phi \mapsto \psi`.
        * ``ShiftFrequency``, with a frequency value :math:`\mu` at time-step :math:`k`, updates
          :math:`\phi_a \mapsto \phi_a - \mu k dt` and :math:`\Delta\nu \mapsto \Delta\nu + \mu`.
          The simultaneous shifting of both :math:`\Delta\nu` and :math:`\phi_a` ensures that the
          carrier wave, as a combination of the analog and digital components, is continuous across
          ``ShiftFrequency`` instructions (up to the sampling rate :math:`dt`).
        * ``SetFrequency``, with a frequency value :math:`\mu` at time-step :math:`k`, updates
          :math:`\phi_a \mapsto \phi_a - (\mu - (\Delta\nu + \nu)) k dt` and
          :math:`\Delta\nu \mapsto \mu - \nu`, where :math:`\nu` is the analog carrier frequency.
          Similarly to ``ShiftFrequency``, the shift rule for :math:`\phi_a` is defined to maintain
          carrier wave continuity.

        If, at any sample point :math:`k`, :math:`\Delta\nu(k)` is larger than the Nyquist sampling
        rate given by ``dt``, a warning will be raised.

        Args:
            schedule: The schedule to represent in terms of signals. Instances of
                :class:`~qiskit.pulse.ScheduleBlock` must first be converted to
                :class:`~qiskit.pulse.Schedule` using the
                :func:`~qiskit.pulse.transforms.block_to_schedule` function in Qiskit Pulse.

        Returns:
            A list of :class:`.DiscreteSignal` instances.
        """

        signals, phases, frequency_shifts, phase_accumulations = {}, {}, {}, {}

        if self._channels is not None:
            schedule = schedule.filter(channels=[self._get_channel(ch) for ch in self._channels])

        for chan in schedule.channels:
            phases[chan.name] = 0.0
            frequency_shifts[chan.name] = 0.0
            phase_accumulations[chan.name] = 0.0

            carrier_freq = self._carriers.get(chan.name, 0.0)

            signals[chan.name] = DiscreteSignal(
                samples=[],
                dt=self._dt,
                name=chan.name,
                carrier_freq=carrier_freq,
            )

        for start_sample, inst in schedule.instructions:
            # get channel name if instruction has it
            chan = inst.channel.name if hasattr(inst, "channel") else None

            if isinstance(inst, Play):
                # get the instruction samples
                inst_samples = None
                if isinstance(inst.pulse, Waveform):
                    inst_samples = inst.pulse.samples
                else:
                    inst_samples = get_samples(inst.pulse)

                # build sample array to append to signal
                times = self._dt * (start_sample + np.arange(len(inst_samples)))
                samples = inst_samples * unp.exp(
                    2.0j * np.pi * frequency_shifts[chan] * times
                    + 1.0j * phases[chan]
                    + 2.0j * np.pi * phase_accumulations[chan]
                )
                signals[chan].add_samples(start_sample, samples)

            if isinstance(inst, ShiftPhase):
                phases[chan] += inst.phase

            if isinstance(inst, SetPhase):
                phases[chan] = inst.phase

            if isinstance(inst, ShiftFrequency):
                frequency_shifts[chan] = frequency_shifts[chan] + inst.frequency
                phase_accumulations[chan] = (
                    phase_accumulations[chan] - inst.frequency * start_sample * self._dt
                )
                _nyquist_warn(frequency_shifts[chan], self._dt, chan)

            if isinstance(inst, SetFrequency):
                phase_accumulations[chan] = phase_accumulations[chan] - (
                    (inst.frequency - (frequency_shifts[chan] + signals[chan].carrier_freq))
                    * start_sample
                    * self._dt
                )
                frequency_shifts[chan] = inst.frequency - signals[chan].carrier_freq
                _nyquist_warn(frequency_shifts[chan], self._dt, chan)

        # ensure all signals have the same number of samples
        max_duration = 0
        for sig in signals.values():
            max_duration = max(max_duration, sig.duration)

        for sig in signals.values():
            if sig.duration < max_duration:
                sig.add_samples(
                    start_sample=sig.duration,
                    samples=np.zeros(max_duration - sig.duration, dtype=complex),
                )

        # filter the channels
        if self._channels is None:
            return list(signals.values())

        return_signals = []
        for chan_name in self._channels:
            signal = signals.get(
                chan_name, DiscreteSignal(samples=[], dt=self._dt, name=chan_name, carrier_freq=0.0)
            )

            return_signals.append(signal)
        return return_signals

    @staticmethod
    def get_awg_signals(
        signals: List[DiscreteSignal], if_modulation: float
    ) -> List[DiscreteSignal]:
        r"""
        Create signals that correspond to the output ports of an Arbitrary Waveform Generator to be
        used with IQ mixers. For each signal in the list the number of signals is double to create
        the I and Q components. The I and Q signals represent the real and imaginary parts,
        respectively, of

        .. math::
            \Omega(t) e^{i \omega_{if} t}

        where :math:`\Omega` is the complex-valued pulse envelope and :math:`\omega_{if}` is the
        intermediate frequency.

        Args:
            signals: A list of signals for which to create I and Q.
            if_modulation: The intermediate frequency with which the AWG modulates the pulse
                envelopes.

        Returns:
            iq signals: A list of signals which is twice as long as the input list of signals. For
                each input signal get_awg_signals returns two
        """
        new_signals = []

        for sig in signals:
            new_freq = sig.carrier_freq + if_modulation

            samples_i = sig.samples
            samples_q = unp.imag(samples_i) - 1.0j * unp.real(samples_i)

            sig_i = DiscreteSignal(
                sig.dt,
                samples_i,
                sig.start_time,
                new_freq,
                sig.phase,
                sig.name + "_i",
            )
            sig_q = DiscreteSignal(
                sig.dt,
                samples_q,
                sig.start_time,
                new_freq,
                sig.phase,
                sig.name + "_q",
            )

            new_signals += [sig_i, sig_q]

        return new_signals

    def _get_channel(self, channel_name: str):
        """Return the channel corresponding to the given name."""

        try:
            prefix = channel_name[0]
            index = int(channel_name[1:])

            if prefix == "d":
                return DriveChannel(index)

            if prefix == "m":
                return MeasureChannel(index)

            if prefix == "u":
                return ControlChannel(index)

            if prefix == "a":
                return AcquireChannel(index)

            raise QiskitError(
                f"Unsupported channel name {channel_name} in {self.__class__.__name__}"
            )

        except (KeyError, IndexError, ValueError) as error:
            raise QiskitError(
                f"Invalid channel name {channel_name} given to {self.__class__.__name__}."
            ) from error


def get_samples(pulse: SymbolicPulse) -> ArrayLike:
    """Return samples filled according to the formula that the pulse represents and the parameter
    values it contains.

    Args:
        pulse: SymbolicPulse class.
    Returns:
        Samples of the pulse.
    Raises:
        PulseError: When parameters are not assigned.
        PulseError: When expression for pulse envelope is not assigned.
        PulseError: When a free symbol value is not defined in the pulse instance parameters.
    """
    envelope = pulse.envelope
    pulse_params = pulse.parameters
    if pulse.is_parameterized():
        raise PulseError("Unassigned parameter exists. All parameters must be assigned.")

    if envelope is None:
        raise PulseError("Pulse envelope expression is not assigned.")

    args = []
    try:
        backend = (
            "jax" if any(isinstance(v, jax.core.Tracer) for v in pulse_params.values()) else "numpy"
        )
    except (ImportError, NameError):
        backend = "numpy"
    for symbol in sorted(envelope.free_symbols, key=lambda s: s.name):
        if symbol.name == "t":
            times = unp.arange(0, pulse_params["duration"]) + 1 / 2
            args.insert(0, times)
            continue
        try:
            args.append(pulse_params[symbol.name])
        except KeyError as ex:
            raise PulseError(
                f"Pulse parameter '{symbol.name}' is not defined for this instance. "
                "Please check your waveform expression is correct."
            ) from ex
    return _lru_cache_expr(
        envelope,
        backend,
    )(*args)


@functools.lru_cache(maxsize=None)
def _lru_cache_expr(expr: sym.Expr, backend) -> Callable:
    """A helper function to get lambdified expression.

    Args:
        expr: Symbolic expression to evaluate.
        backend: Array backend.
    Returns:
        lambdified expression.
    """
    params = []
    for param in sorted(expr.free_symbols, key=lambda s: s.name):
        if param.name == "t":
            params.insert(0, param)
            continue
        params.append(param)
    return sym.lambdify(params, expr, modules=backend)


def _nyquist_warn(frequency_shift: ArrayLike, dt: float, channel: str):
    """Raise a warning if the frequency shift is above the Nyquist frequency given by ``dt``."""
    if (
        isinstance(frequency_shift, (int, float, list, np.ndarray))
        or not isinstance(jnp.array(0), jax.core.Tracer)
    ) and np.abs(frequency_shift) > 0.5 / dt:
        warn(
            "Due to SetFrequency and ShiftFrequency instructions, the digital carrier frequency "
            f"of channel {channel} is larger than the Nyquist frequency of the envelope sample "
            "size dt. As shifts of the frequency from the analog frequency are handled digitally, "
            "this will result in aliasing effects."
        )



# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=invalid-name

"""
Wrapper for calling scipy.integrate.solve_ivp.
"""

from typing import Callable, Union, Optional

import numpy as np
from scipy.integrate import solve_ivp, OdeSolver
from scipy.integrate._ivp.ivp import OdeResult

from qiskit import QiskitError
from qiskit_dynamics.arraylias import ArrayLike

# Supported scipy ODE methods
COMPLEX_METHODS = ["RK45", "RK23", "BDF", "DOP853"]
REAL_METHODS = ["LSODA", "Radau"]
SOLVE_IVP_METHODS = COMPLEX_METHODS + REAL_METHODS


def scipy_solve_ivp(
    rhs: Callable,
    t_span: ArrayLike,
    y0: ArrayLike,
    method: Union[str, OdeSolver],
    t_eval: Optional[ArrayLike] = None,
    **kwargs,
):
    """Routine for calling `scipy.integrate.solve_ivp`.

    Args:
        rhs: Callable of the form :math:`f(t, y)`.
        t_span: Interval to solve over.
        y0: Initial state.
        method: Solver method.
        t_eval: Points at which to evaluate the solution.
        **kwargs: Optional arguments to be passed to ``solve_ivp``.

    Returns:
        OdeResult: results object

    Raises:
        QiskitError: If unsupported kwarg present.
    """

    if kwargs.get("dense_output", False) is True:
        raise QiskitError("dense_output not supported for solve_ivp.")

    y_shape = y0.shape

    # flatten y0 and rhs
    y0 = y0.flatten()
    rhs = flat_rhs(rhs, y_shape)

    # Check if solver is real only
    # TODO: Also check if model or y0 are complex
    #       if they are both real we don't need to embed.
    embed_real = method in REAL_METHODS
    if embed_real:
        rhs = real_rhs(rhs)
        y0 = c2r(y0)

    results = solve_ivp(rhs, t_span=t_span, y0=y0.data, t_eval=t_eval, method=method, **kwargs)
    if embed_real:
        results.y = r2c(results.y)

    # convert to the standardized results format
    # solve_ivp returns the states as a 2d array with columns being the states
    results.y = results.y.transpose()
    results.y = np.array([y.reshape(y_shape) for y in results.y])

    return OdeResult(**dict(results))


def flat_rhs(rhs, shape):
    """Convert an RHS with arbitrary state shape into one that is 1d."""

    def _flat_rhs(t, y):
        return rhs(t, y.reshape(shape)).flatten()

    return _flat_rhs


def real_rhs(rhs):
    """Convert complex RHS to real RHS function"""

    def _real_rhs(t, y):
        return c2r(rhs(t, r2c(y)))

    return _real_rhs


def c2r(arr):
    """Convert complex array to a real array"""
    return np.concatenate([np.real(arr), np.imag(arr)])


def r2c(arr):
    """Convert a real array to a complex array"""
    size = arr.shape[0] // 2
    return arr[:size] + 1j * arr[size:]



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=invalid-name

"""
Wrapper for diffrax solvers
"""

from typing import Callable, Optional
from scipy.integrate._ivp.ivp import OdeResult
from qiskit import QiskitError

from qiskit_dynamics.arraylias import ArrayLike, requires_array_library

try:
    import jax.numpy as jnp
except ImportError:
    pass


@requires_array_library("jax")
def diffrax_solver(
    rhs: Callable,
    t_span: ArrayLike,
    y0: ArrayLike,
    method: "AbstractSolver",
    t_eval: Optional[ArrayLike] = None,
    **kwargs,
):
    """Routine for calling ``diffrax.diffeqsolve``

    Args:
        rhs: Callable of the form :math:`f(t, y)`.
        t_span: Interval to solve over.
        y0: Initial state.
        method: Which diffeq solving method to use.
        t_eval: Optional list of time points at which to return the solution.
        **kwargs: Optional arguments to be passed to ``diffeqsolve``.

    Returns:
        OdeResult: Results object.

    Raises:
        QiskitError: Passing both `SaveAt` argument and `t_eval` argument.
    """

    from diffrax import ODETerm, SaveAt
    from diffrax import diffeqsolve

    # convert rhs and y0 to real
    rhs = real_rhs(rhs)
    y0 = c2r(y0)

    term = ODETerm(lambda t, y, _: rhs(t.real, y))

    if "saveat" in kwargs and t_eval is not None:
        raise QiskitError(
            """Only one of t_eval or saveat can be passed when using
            a diffrax solver, but both were specified."""
        )

    if t_eval is not None:
        kwargs["saveat"] = SaveAt(ts=t_eval)

    results = diffeqsolve(
        term,
        solver=method,
        t0=t_span[0],
        t1=t_span[-1],
        dt0=None,
        y0=jnp.array(y0, dtype=float),
        **kwargs,
    )

    sol_dict = vars(results)
    ys = sol_dict.pop("ys")
    ts = sol_dict.pop("ts")

    ys = jnp.swapaxes(r2c(jnp.swapaxes(ys, 0, 1)), 0, 1)

    results_out = OdeResult(t=ts, y=jnp.array(ys, dtype=complex), **sol_dict)

    return results_out


def real_rhs(rhs):
    """Convert complex RHS to real RHS function"""

    def _real_rhs(t, y):
        return c2r(rhs(t, r2c(y)))

    return _real_rhs


def c2r(arr):
    """Convert complex array to a real array"""
    return jnp.concatenate([jnp.real(arr), jnp.imag(arr)])


def r2c(arr):
    """Convert a real array to a complex array"""
    size = arr.shape[0] // 2
    return arr[:size] + 1j * arr[size:]



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

r"""
========================================
Backend (:mod:`qiskit_dynamics.backend`)
========================================

.. currentmodule:: qiskit_dynamics.backend

This module contains the :class:`.DynamicsBackend` class, which provides a
:class:`~qiskit.providers.backend.BackendV2` interface for running pulse-level simulations with
Qiskit Dynamics. The :class:`.DynamicsBackend` can directly simulate :class:`~qiskit.pulse.Schedule`
and :class:`~qiskit.pulse.ScheduleBlock` instances, and can also be configured to simulate
:class:`~qiskit.circuit.QuantumCircuit`\s at the pulse-level via circuit to pulse transpilation.

This module also exposes some functions utilized by :class:`.DynamicsBackend` that may be of use to
experienced users. The function :func:`.default_experiment_result_function` is the default method by
which results are computed and returned to the user after the underlying differential equation is
solved. This function can be overridden with a custom user-defined function by setting the
``experiment_result_function`` option of :class:`.DynamicsBackend`.


Classes and functions
=====================

.. autosummary::
   :toctree: ../stubs/

   DynamicsBackend
   default_experiment_result_function
"""

from .dynamics_backend import DynamicsBackend, default_experiment_result_function



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=invalid-name, inconsistent-return-statements

"""Functions for performing the Rotating Wave Approximation
on Model classes."""


from typing import List, Optional, Union
import numpy as np

from qiskit_dynamics import DYNAMICS_NUMPY as unp
from qiskit_dynamics.arraylias.alias import ArrayLike, _to_dense, _to_dense_list
from qiskit_dynamics.models import (
    BaseGeneratorModel,
    GeneratorModel,
    HamiltonianModel,
    LindbladModel,
    RotatingFrame,
)
from qiskit_dynamics.signals import SignalSum, Signal, SignalList


def rotating_wave_approximation(
    model: BaseGeneratorModel, cutoff_freq: float, return_signal_map: Optional[bool] = False
) -> BaseGeneratorModel:
    r"""Construct a new model by performing the rotating wave approximation with a given cutoff
    frequency. The outputs of this function can be used in JAX-transformable functions, however this
    function itself cannot (see below).

    Performs elementwise rotating wave approximation (RWA) with cutoff frequency ``cutoff_freq`` on
    each operator in a model, returning a new model. The new model contains a modified list of
    signal coefficients, and setting the optional argument ``return_signal_map=True`` results in
    the additional return of the function ``f`` which maps the signals of the input model to those
    of the output RWA model, such that the code blocks:

    .. code-block:: python

        model.signals = new_signals
        rwa_model = rotating_wave_approximation(model, cutoff_freq)

    and

    .. code-block:: python

        rwa_model, f = rotating_wave_approximation(model, cutoff_freq, return_signal_map=True)
        rwa_model.signals = f(new_signals)

    result in an ``rwa_model`` with the same updated signals.

    .. note::
        The ``rotating_wave_approximation`` function itself cannot be included in a function
        to-be JAX-transformed, however the resulting model and ``signal_map`` can. For example,
        the following function is **not** JAX-transformable:

        .. code-block:: python

            def function_with_rwa(t):
                operators = ...
                signals = ...
                model = GeneratorModel(operators=operators, signals=signals)
                rwa_model = rotating_wave_approximation(model, cutoff_freq)
                return rwa_model(t)

        Whereas, defining:

        .. code-block:: python

            rwa_model, signal_map = rotating_wave_approximation(
                model,
                cutoff_freq,
                return_signal_map=True
            )

        The following function **is** JAX-transformable:

        .. code-block:: python

            def jax_transformable_func(t):
                rwa_model.signals = signal_map(new_signals)
                return rwa_model(t)

        In this way, the outputs of ``rotating_wave_approximation`` can be used in JAX-transformable
        functions, however ``rotating_wave_approximation`` itself cannot.


    We now describe the formalism. When considering :math:`s_i(t) e^{-tF}G_ie^{tF}`, in the basis in
    which :math:`F` is diagonal, the :math:`(j, k)` element of :math:`G_i` has effective frequency
    :math:`\tilde\nu_{ijk}^\pm = \pm\nu_i + Im[-d_j+d_k]/2\pi`, where the :math:`\pm\nu_i` comes
    from expressing :math:`s_i(t) = Re[a_i(t)e^{2\pi i\nu_i t}] = a_i(t)e^{i(2\pi\nu_i t+\phi_i)}/2
    + c.c.` and the other term comes from the rotating frame. Define :math:`G_i^\pm` the matrix
    whose entries :math:`(G_i^\pm)_{jk}` are the entries of :math:`G_i` s.t.
    :math:`|\nu_{ijk}^\pm|<\nu_*` for some cutoff frequency :math:`\nu_*`. Then, after the RWA, we
    may write

    .. math::
        s_i(t)G_i \to G_i^+ a_ie^{i(2\pi \nu_i t+\phi_i)}/2
                        + G_i^- \overline{a_i}e^{-i(2\pi \nu_i t+\phi_i)}/2.

    When we regroup these to use only the real components of the signal, we find that

    .. math::
        s_i(t)G_i \to  s_i(t)(G_i^+ + G_i^-)/2 + s_i'(t)(iG_i^+-iG_i^-)

    where :math:`s_i'(t)` is a signal with the same frequency and amplitude as :math:`s_i`, but with
    a phase shift of :math:`\phi_i - \pi/2`.

    Args:
        model: The model to approximate.
        cutoff_freq: The cutoff frequency for the approximation.
        return_signal_map: Whether to also return the function for mapping pre-RWA signals to
            post-RWA signals.

    Returns:
        :class:`GeneratorModel` with twice as many terms, and, if ``return_signal_map``,
        also the function ``f``.

    Raises:
        ValueError: If the model has no signals.
    """

    n = model.dim

    frame_freqs = None
    if model.rotating_frame is None or model.rotating_frame.frame_diag is None:
        frame_freqs = np.zeros((n, n), dtype=complex)
    else:
        diag = model.rotating_frame.frame_diag
        diff_matrix = np.broadcast_to(diag, (n, n)) - np.broadcast_to(diag, (n, n)).T
        frame_freqs = diff_matrix.imag / (2 * np.pi)

    if model.rotating_frame.frame_diag is not None:
        frame_shift = np.diag(model.rotating_frame.frame_diag)
        if isinstance(model, (HamiltonianModel, LindbladModel)):
            frame_shift = 1j * frame_shift
    else:
        frame_shift = np.zeros((n, n), dtype=complex)

    if isinstance(model, GeneratorModel):
        if model.signals is None and model.operators is not None:
            raise ValueError("Model must have nontrivial signals to perform the RWA.")

        cur_drift = _to_dense(model._operator_collection.static_operator)
        if isinstance(model, HamiltonianModel) and cur_drift is not None:
            cur_drift = 1j * cur_drift

        if cur_drift is not None:
            cur_drift = cur_drift + frame_shift
            rwa_drift = cur_drift * (abs(frame_freqs) < cutoff_freq).astype(int)
            rwa_drift = model.rotating_frame.operator_out_of_frame_basis(rwa_drift)
        else:
            rwa_drift = None

        operators = _to_dense_list(model._operator_collection.operators)
        if isinstance(model, HamiltonianModel) and operators is not None:
            operators = 1j * operators

        rwa_operators = get_rwa_operators(
            operators,
            model.signals,
            model.rotating_frame,
            frame_freqs,
            cutoff_freq,
        )
        rwa_signals = get_rwa_signals(model.signals)

        # works for both GeneratorModel and HamiltonianModel
        rwa_model = model.__class__(
            static_operator=rwa_drift,
            operators=rwa_operators,
            signals=rwa_signals,
            rotating_frame=model.rotating_frame,
            in_frame_basis=model.in_frame_basis,
            array_library=model.array_library,
        )
        if return_signal_map:
            return rwa_model, get_rwa_signals
        return rwa_model

    elif isinstance(model, LindbladModel):
        if model.signals[0] is None and model.hamiltonian_operators is not None:
            raise ValueError("Model must have nontrivial Hamiltonian signals to perform the RWA.")

        if model.signals[1] is None and model.dissipator_operators is not None:
            raise ValueError("Model must have nontrivial dissipator signals to perform the RWA.")

        # static hamiltonian part
        cur_drift = _to_dense(model._operator_collection.static_hamiltonian) + frame_shift
        rwa_drift = cur_drift * (abs(frame_freqs) < cutoff_freq).astype(int)
        rwa_drift = model.rotating_frame.operator_out_of_frame_basis(rwa_drift)

        # static dissipator part
        cur_static_dis = _to_dense_list(model._operator_collection.static_dissipators)
        rwa_static_dis = None
        if cur_static_dis is not None:
            rwa_static_dis = []
            for op in cur_static_dis:
                op = unp.asarray(op)
                rwa_op = op * (abs(frame_freqs) < cutoff_freq).astype(int)
                rwa_op = model.rotating_frame.operator_out_of_frame_basis(rwa_op)
                rwa_static_dis.append(rwa_op)

        cur_ham_ops = _to_dense_list(model._operator_collection.hamiltonian_operators)
        cur_dis_ops = _to_dense_list(model._operator_collection.dissipator_operators)

        cur_ham_sig, cur_dis_sig = model.signals

        rwa_ham_ops = get_rwa_operators(
            cur_ham_ops, cur_ham_sig, model.rotating_frame, frame_freqs, cutoff_freq
        )
        rwa_ham_sig = get_rwa_signals(cur_ham_sig)

        rwa_dis_ops = get_rwa_operators(
            cur_dis_ops, cur_dis_sig, model.rotating_frame, frame_freqs, cutoff_freq
        )

        rwa_dis_sig = get_rwa_signals(cur_dis_sig)

        rwa_model = LindbladModel(
            static_hamiltonian=rwa_drift,
            hamiltonian_operators=rwa_ham_ops,
            hamiltonian_signals=rwa_ham_sig,
            static_dissipators=rwa_static_dis,
            dissipator_operators=rwa_dis_ops,
            dissipator_signals=rwa_dis_sig,
            rotating_frame=model.rotating_frame,
            in_frame_basis=model.in_frame_basis,
            array_library=model.array_library,
            vectorized=model.vectorized,
        )

        if return_signal_map:
            return rwa_model, lambda a: (get_rwa_signals(a[0]), get_rwa_signals(a[1]))
        return rwa_model


def get_rwa_operators(
    current_ops: ArrayLike,
    current_sigs: SignalList,
    rotating_frame: RotatingFrame,
    frame_freqs: ArrayLike,
    cutoff_freq: float,
) -> ArrayLike:
    r"""Given a set of operators as a ``(k,n,n)`` array, a set of frequencies
    :math:`\operatorname{frame_freqs}_{jk} = \operatorname{Im}[-d_j+d_k]` where :math:`d_i` the
    :math:`i^{th}` eigenlvalue of the frame operator :math:`F`, the current signals of a model, and
    a cutoff frequency, returns the new operators and signals that should be passed to create a new
    Model class after the RWA.

    Args:
        current_ops: The current operator list, a ``(k,n,n)`` array.
        current_sigs: ``(k,)`` length :class:`SignalList`.
        rotating_frame: The current :class:`~RotatingFrame` object of the pre-RWA model
        frame_freqs: The effective frequencies of different matrix elements due to the conjugation
            by :math:`e^{\pm Ft}` in the rotating frame.
        cutoff_freq: The maximum frequency allowed under the RWA.

    Returns:
        SignaLList: ``(2k,n,n)`` array of new operators post RWA.
    """
    if current_ops is None:
        return None

    current_sigs = current_sigs.flatten()
    carrier_freqs = np.zeros(len(current_ops))

    for i, sig_sum in enumerate(current_sigs.components):
        sig = sig_sum.components[0]
        carrier_freqs[i] = sig.carrier_freq

    num_components = len(carrier_freqs)
    n = current_ops[0].shape[-1]

    frame_freqs = np.broadcast_to(frame_freqs, (num_components, n, n))
    carrier_freqs = carrier_freqs.reshape((num_components, 1, 1))

    pos_pass = np.abs(carrier_freqs + frame_freqs) < cutoff_freq
    pos_terms = current_ops * pos_pass.astype(int)  # G_i^+

    neg_pass = np.abs(-carrier_freqs + frame_freqs) < cutoff_freq
    neg_terms = current_ops * neg_pass.astype(int)  # G_i^-

    real_component = pos_terms / 2 + neg_terms / 2
    imag_component = 1j * pos_terms / 2 - 1j * neg_terms / 2

    return rotating_frame.operator_out_of_frame_basis(
        np.append(real_component, imag_component, axis=0)
    )


def get_rwa_signals(curr_signal_list: Union[List[Signal], SignalList]) -> SignalList:
    """Helper function that converts pre-RWA signals to post-RWA signals.

    Args:
        curr_signal_list: The pre-RWA signals.

    Returns:
        The post-RWA signals.
    """
    if curr_signal_list is None:
        return curr_signal_list

    real_signal_components = []
    imag_signal_components = []

    if not isinstance(curr_signal_list, SignalList):
        curr_signal_list = SignalList(curr_signal_list)

    curr_signal_list = curr_signal_list.flatten()

    for sig_sum in curr_signal_list.components:
        sig = sig_sum.components[0]
        real_signal_components.append(sig)
        imag_signal_components.append(
            SignalSum(Signal(sig.envelope, sig.carrier_freq, sig.phase - np.pi / 2))
        )

    return SignalList(real_signal_components + imag_signal_components)
